#!/usr/bin/env python
import SEP.util
import sys,string,re,commands,os
import SEP.operator
import SEP.parfile
import SEP.parjob
import SEP.par
import SEP.sep_vector
import SEP.vector



class wei_operator(SEP.operator.oc_par):
  """Class that treats the migration as an operator"""
  def __init__(self,prog,pars,**kw):
     """Initialize a wei operator

         prog    - the name of the program
         pars    - the parameters associated with the migration
         kw      - the parameters associated with the parallel job

         Parameters specific to the parallel job can be specified
           as extra arguments to the init (**kw list) or with
           par_ prefix in pars
        
         All parameters that aren't of the form par_ are passed
         to both the parallel_file creation and to the wei program


         Some of the parallel job parameters (check SEP.pars.parjob for more):
       
         mfile       - the machine file for the parallel job ("mfile")
         restart     - whether or not we are restarting (None [false])
         logidr      - the directory to write the logfiles ("./wei_log")
         name        - the name for the parallel job
         restart_com - the restart command (restart=y)
         verb        - verbosity level


         Some parameters specific to this wei object:

         nfs_image      -If true the image is written to shared files on NFS disk
         nmax_procs     -nfs image the number of processes that share image file
         share_datapath -the datapath for the shared files
         nsect          - the number of sections to break the frequency axis into

         Parameter for wei library (many of these are created by inherited classes) :

         D             - the frequency file (mx,my,hx,hy,w)
         R             - the image file     (mx,my,hx,hw,z)
         S             - the slowness/velocity file (mx,my,z)
         slow_choice   - the slowness to use for the downward continuation (nref,2,nz)
         velocity      - whether or not we have velocity (y) or slowness (n),default
         image_real    - whether or not to make the model real (y) or complex (n),default
         wave_out      - the file to write the wavefield at the bottom (default None)
          
     """
     pars.add_dictionary(kw,"par_")
     par_par=pars.extract_del_pars("par_")

     par_par.default_val("mfile","mfile")
     par_par.default_val("njobs",2000)
     restart=par_par.logic("restart")
     logdir=par_par.param("logdir","wei_log")
     name=par_par.param("name","wei_run")
     par_par.default_val("restart_com","restart=y")

     self.pars=pars
     self.check_wavespace()
     slowS,schoice=self.slow_sep()
     isect_dict,sect_par,order=self._build_dicts()
     parfiles={}
     self.program=prog
     parfiles["schoice"]=SEP.parfile.copy(name=schoice.name,tag="schoice=",
        isect_dict=isect_dict,usage="INPUT",type="COPY",logdir=logdir)
     parfiles["slow"]=SEP.parfile.copy(name=slowS.name,tag="S=",
        isect_dict=isect_dict,usage="INPUT",type="COPY",logdir=logdir)
     parfiles["range"]=SEP.parfile.distribute(name="data",tag="D=",
        isect_dict=isect_dict,space_only=1,dff_axis=5,logdir=logdir,
       nblock=[len(isect_dict.keys())])
     nfs=None 
     if pars.logic("nfs_image",delete=1): nfs=1

     if nfs:
       sz=int(self.pars.param("az__n"))* int(self.pars.param("amy_n"))* int(self.pars.param("amx_n"))* int(self.pars.param("ahx_n"))* int(self.pars.param("ahy_n"))*4
       big=""
       if sz > 1024*1024*2000: 
         nfiles=sz/1024/1024/2000
         big="nfiles=%d maxsize=2000"%(nfiles+1)
       nmax=pars.param("nmax_procs",8,delete=1)
       dpath=pars.param("share_datapath",None,error=1,delete=1)
       parfiles["domain"]=SEP.parfile.copy_share(name="model",tag="R=",
          logdir=logdir,extra_args=big,
         isect_dict=isect_dict,space_only=1,type="COPY",
         reuse_par="add=y",nmax_procs=nmax, share_datapath=dpath)
     else:
       parfiles["domain"]=SEP.parfile.copy(name="model",tag="R=",logdir=logdir,
          isect_dict=isect_dict,space_only=1,type="COPY",reuse_par="add=y")
     if pars.par("wave_out")[1]==1:
       parfiles["wave_out"]=SEP.parfile.distribute(
         name=pars.param("wave_out",delete=1),tag="wave_out=",
        isect_dict=isect_dict,dff_axis=5,logdir=logdir,
        usage="OUTPUT", nblock=[len(isect_dict.keys())])
     model=self.create_model("mod_space")
     data=self.create_data("dat_space")

     if nfs:
       parjob=SEP.parjob.parjob_nfs(files=parfiles,sect_pars=sect_par,
        program=self.program, order=order, **par_par.return_dict())
     else:
       parjob=SEP.parjob.parjob(files=parfiles,sect_pars=sect_par,
        program=self.program, order=order, **par_par.return_dict())

     if not name: name=self.program
     SEP.operator.oc_par.__init__(self,name,
       model,data,parjob,pars=pars,model_tag="R=",data_tag="D=",restart=restart)

     self.model_check={}
     self.data_check={}
     self.verb=pars.logic("verb")


  def _build_dicts(self):
    """Build the parameter and section databases """
    isect_dict={}
    par_dict={}
    nw=int(self.pars.par("aw__n")[0])
    ow=float(self.pars.par("aw__o")[0])
    dw=float(self.pars.par("aw__d")[0])
    nsect=int(self.pars.param("nsect",nw))
    order=[]
    nb=int(nw/nsect)
    nextra=nw-nsect*nb
    itot=0
    for i in range(nsect): 
       key=str(i)
       order.append(key)
       isect_dict[key]=i
       par_dict[key]=SEP.par.sep_pars(name=str(i))
       o=ow+dw*itot
       n=nb
       if i < nextra: n=n+1
       par_dict[key].add_param("aw__n",n)
       par_dict[key].add_param("aw__o",o)
       par_dict[key].add_param("aw__d",dw)
       itot=itot+n
    return isect_dict,par_dict,order

  def add_params(self,pars):
    """Add and check parameters"""
    SEP.operator.oc.add_params(self,pars)
    self.check_slow(slow)

  def slow_sep(self):
    """Check to make sure the slowness make sense

       The slowness file needs to be 3 dimensional regular dataset that
       has the same depth sampling as the image

    """
    val,err = self.pars.par("S")
    if err !=1: SEP.util.err("Expecting slowness in the form S=file.H")
    slow=SEP.sepfile.sep_file(file=val)
    val,err = self.pars.par("schoice")
    if err !=1: SEP.util.err("Expecting ref slowness in schoice=file.H")
    schoice=SEP.sepfile.sep_file(file=val)

    if schoice.headers or schoice.grid:
      SEP.util.err("Expecting schoice to be a regular SEPlib file")

    if slow.headers or slow.grid:
      SEP.util.err("Expecting slowness to be a regular SEPlib file")

    if schoice.ndims()!=3: SEP.util.err("Expecting schoice to be a 3-D file nref,2,nz")
    if slow.ndims()!=3: SEP.util.err("Expecting slowness to be a 3-D file x,y,z")
    if slow.history_par("n1")[0] > 1:
       b,e=slow.axis_min_max(1)
       b1=self.pars.param("wave_x_min",delete=1)
       e1=self.pars.param("wave_x_max",delete=1)
       if b >b1: SEP.util.msg("Warning: Slowness axis 1 will be extended (neg) by %f "%(b-b1))
       if e <e1: SEP.util.msg("Warning: Slowness axis 1 will be extended (pos) by %f "%(e1-e))
    n3_s,oz,dz,label,unit=schoice.history.axis(3)
    n3_c=slow.history_par("n3",error=1)[0]
    if int(n3_s) != int(n3_c): 
       SEP.util.err("Expecting slowness and schoice to have identical n3")

    if n3_s!=int(self.pars.par("az__n")[0]):
      SEP.util.err("Number of samples slowness (axis 3) not the same")
    if oz!=float(self.pars.par("az__o")[0]):
      SEP.util.err("First sample of slowness (axis 3) not the same")
    if dz!=float(self.pars.par("az__d")[0]):
      SEP.util.err("Sampling of slowness (axis 3) not the same")


    if slow.history_par("n2")[0] > 1:
       b,e=slow.axis_min_max(2)
       b1=self.pars.param("wave_y_min",delete=1)
       e1=self.pars.param("wave_y_max",delete=1)
       if b >b1: SEP.util.msg("Warning: Slowness axis 2 will be extended (neg) by %f "%(b-b1))
       if e <e1: SEP.util.msg("Warning: Slowness axis 2 will be extended (pos) by %f "%(e1-e))

       return slow,schoice

  def check_model(self,model):
    """Check to see if the model makes sense
        
       The model needs to be 5-D mx,my,hx,hy,z
        
       The model description must correspond with amx,amy,ahx,ahy,az_

    """
    if self.model_check.has_key(model.name): return
    if model.headers or model.grid:
      SEP.util.err("Model %s should be a regular dataset"%model.name)
    self._axis_check("Model",model,1,"mx")
    self._axis_check("Model",model,2,"my")
    self._axis_check("Model",model,3,"hx")
    self._axis_check("Model",model,4,"hy")
    self._axis_check("Model",model,5,"z_")
    self.model_check[model.name]=1

  def _axis_check(self,name,sep,axis,pp):
    """Check to see if the axis makes sense

       name - the axis name
       sep  - the sepfile
       axis - the axis number
       pp   - the prefix for the axis

       Example:
       
       mig.__axis_check("midpoint x",D,1,"mx")

    """ 
    n,o,d,label,unit=sep.axis(axis)
    if n>1: 
      n2=self.pars.par("a%s_n"%pp)[0]
      o2=self.pars.par("a%s_o"%pp)[0]
      d2=self.pars.par("a%s_d"%pp)[0]
      SEP.sepfile.axis_same("%s(%s) axis %d"%(name,sep.name,axis),n,o,d,pp,n2,o2,d2)

  def check_data(self,data):
    """Check to see if the data makes sense

       The data needs to be 5-D mx,my,hx,hy,w
        
       The data description must correspond with amx,amy,ahx,ahy,aw_

    """
    if self.data_check.has_key(data.name): return
    if data.headers or data.grid:
      SEP.util.err("Data %s should be a regular dataset"%data.name)
    self._axis_check("Data",data,1,"mx")
    self._axis_check("Data",data,2,"my")
    self._axis_check("Data",data,3,"hx")
    self._axis_check("Data",data,4,"hy")
    self._axis_check("Data",data,5,"w_")
    self.data_check[data.name]=1
    


  def check_wavespace(self):
    """Check to make sure wavespace is described

      Check to make sure the following are in the parameters

      amx_n,amy_n,ahx_n,ahy_n,az__n,aw__n
      amx_o,amy_o,ahx_o,ahy_o,az__o,aw__o
      amx_d,amy_d,ahx_d,ahy_d,az__d,aw__d

    """
    nmx,omx,dmx = self.check_wave_axis("mx")
    nmy,omy,dmy = self.check_wave_axis("my")
    nhx,ohx,dhx = self.check_wave_axis("hx")
    nhy,ohy,dhy = self.check_wave_axis("hy")
    nz,oz,dz = self.check_wave_axis("z_")
    nw,ow,dw = self.check_wave_axis("w_")

    bm,em= SEP.sepfile.min_max(nmx,omx,dmx)
    bh,eh= SEP.sepfile.min_max(nhx,ohx,dhx)
    if abs(bh) > abs(eh) : mx=abs(bh)
    else: mx=abs(eh)
    self.pars.add_param("wave_x_min",bm-mx)
    self.pars.add_param("wave_x_max",em+mx)

    bm,em= SEP.sepfile.min_max(nmy,omy,dmy)
    bh,eh= SEP.sepfile.min_max(nhy,ohy,dhy)
    if abs(bh) > abs(eh) : mx=abs(bh)
    else: mx=abs(eh)
    self.pars.add_param("wave_y_min",bm-mx)
    self.pars.add_param("wave_y_max",em+mx)

    z,err=self.pars.par("nzb",1)
    z=int(z)
    if nz/z != int(float(nz)/float(z)): 
      SEP.util.err("nzb(%d) is not a factor of %d "%(z,nz))
    self.pars.add_param("nw_tot",nw)

  def check_wave_axis(self,par,ndef=None,odef=None,ddf=None):
    """Return the n,o,d [with possible defaults for a wavespace axis"""
    n,err=self.pars.par("a%s_n"%par,ndef)
    if err==0: SEP.util.err("Required parameter a%s_n missing"%par)
    o,err=self.pars.par("a%s_o"%par,ndef)
    if err==0: SEP.util.err("Required parameter a%s_o missing"%par)
    d,err=self.pars.par("a%s_d"%par,ndef)
    if err==0: SEP.util.err("Required parameter a%s_d missing"%par)
    return int(n),float(o),float(d)

  def create_model(self,name):
    """Create the model space with the file name 'name'"""
    do_real=self.pars.par("image_real","n")[0]
    if str(do_real)[0]=="y" or  str(do_real)[0]=="1":do_real=1
    else: do_real=None
    return self._create_file(name,["mx","my","hx","hy","z_"],do_real)

  def create_data(self,name):
    """Create the data space with the file name name"""
    return self._create_file(name,["mx","my","hx","hy","w_"])

  def _create_file(self,name,axes,do_real=None):
    """Create  a file using the axis description in axes

       name - the name of the file to create
       axes - the list of axes to use (["mx","my","hx","hy","w_"]
       do_real- wheter of not the data is real

    """
    if do_real: junk=SEP.sep_vector.sfloat(name=name,space_only=1)
    else:       junk=SEP.sep_vector.scmplx(name=name,space_only=1)
    for i in range(len(axes)):
       n,o,d=self.get_wave_axis(axes[i])
       junk.set_axis(i+1,n,o,d,axes[i],"none")
    if do_real:junk.history.add_param("esize",4)
    else:junk.history.add_param("esize",8)
    return junk

  def get_wave_axis(self,par):
    """Return aX_n, aX_o, aX_d"""
    n=self.pars.par("a%s_n"%par)[0] 
    o=self.pars.par("a%s_o"%par)[0] 
    d=self.pars.par("a%s_d"%par)[0] 
    return n,o,d
  
       

class wei(wei_operator):
  """Basic wei object"""
  def __init__(self,pars,prog,**kw):
    """Initialize the basic wei operator 

       pars    - the parameters to use
       prog    - the program to run
       kw      - additional parameters for the parallel job

       See SEP.wei_oper.wei_operator for a more complete descrition of parametters

    """
    wei_operator.__init__(self,prog,pars,**kw)

class wei_ang(wei):
  """Class to do wem with ph gathers"""
  def __init__(self,pars,prog, **kw):
    """Initialize the basic wei operator for angle gathers

       pars    - the parameters to use
       prog    - the program to run
       kw      - additional parameters for the parallel job

       You need to specify in pars
       apx_n,apx_o,apx_d - ph axis description in x
       apy_n,apy_o,apy_d - ph axis description in y

       See SEP.wei_oper.wei_operator for further parameter description


    """
    pars.add_string("ang_par=y")
    wei.__init__(self,pars,prog,**kw)

  def check_model(self,model):
    """Check to see if the model makes sense"""
    if self.model_check.has_key(model.name): return
    if model.headers or model.grid:
      SEP.util.err("Model %s should be a regular dataset"%model.name)
    self._axis_check("Model",model,1,"mx")
    self._axis_check("Model",model,2,"my")
    n,o,d,label,unit=sep.axis(3)
    n2=self.pars.par("apx_n",error=1)[0]
    n3=self.pars.par("ahx_n",error=1)[0]
    if n!=n2 or n!=n3: 
      SEP.util.err("Length of  axis 3 of model, apx_n,ahx_n must be the same")
    n,o,d,label,unit=sep.axis(4)
    n2=self.pars.par("apy_n",error=1)[0]
    n3=self.pars.par("ahy_n",error=1)[0]
    if n!=n2 or n!=n3: 
      SEP.util.err("Length of  axis 3 of model, apy_n,apx_n must be the same")
    self._axis_check("Model",model,5,"z_")
    self.model_check[model.name]=1

  def create_model(self,name):
    """Create the model structure"""
    do_real=self.pars.par("image_real","n")[0]
    if str(do_real)[0]=="y" or  str(do_real)[0]=="1":do_real=1
    else: do_real=None
    return self._create_file(name,["mx","my","px","py","z_"],do_real)


class wei_migration(wei):
  """WEI  migration"""
  def __init__(self,pars,prog,**kw):
    """Initialize wave equation migration

      pars  - the parameters for the migration
      prog  - the program to run
      kw    - additional parameters for the parallel job

      See SEP.wei_oper.wei_operator for further parameter description

      pars must specify az__n,az__o,az__d

    """
    pars.add_string("operation=migration")
    if kw.has_key("restart"): restart=kw["restart"]
    else: restart=None
    wei.__init__(self,pars,prog,**kw)
    mtag,err1=self.pars.par("R")
    dtag,err2=self.pars.par("D")
    if err1==0: SEP.util.err("When doing migration must provide model (R)")
    if err2==0: SEP.util.err("When doing migration must provide data (D)")
    self.data=SEP.sep_vector.scmplx(tag=dtag)
    self.check_data(self.data)
    if self.verb: SEP.util.msg("Initializing/reading a model")
    self.image=self.create_model(mtag)
    if not restart: 
        self.image.write_file()
        self.image.zero()
    self.restart=restart
  def migrate(self):
    """Run a migration"""
    self.run_adjoint(self.image.name,self.data.name,restart=self.restart)

class wei_modeling(wei):
  """WEM  migration"""
  def __init__(self,pars,prog=None,**kw):
    """Initialize wave equation modeling

      pars  - the parameters for the migration
      prog  - the program to run
      kw    - additional parameters for the parallel job

      See SEP.wei_oper.wei_operator for further parameter description

      pars must specify aw__n,aw__o,aw__d

    """
    pars.add_string("operation=modeling")
    wei.__init__(self,pars,prog,**kw)
    mtag,err1=self.pars.par("R")
    dtag,err2=self.pars.par("D")
    if err1==0: SEP.util.err("When doing modeling must provide model (R)")
    if err2==0: SEP.util.err("When doing modeling must provide data (D)")
    model=SEP.sepfile.sep_file(file=mtag)
    esize=model.history_par("esize",4)[0]
    if int(esize)==4: self.image=SEP.sep_vector.sfloat(tag=mtag)
    else: self.image=vector.scmplx(tag=mtag)
    self.image=SEP.sepfile.sep_file(file=mtag)
    self.check_model(model)
    if self.verb: SEP.util.msg("Initializing/reading a data")
    self.data=self.create_data(dtag)
    if kw.has_key("restart"): restart=kw["restart"]
    else: restart=None
    if not restart: self.data.zero()
    self.restart=restart
  def model(self):
    """Run a modelling job"""
    self.run_forward(self.image.name,self.data.name,restart=self.restart)
  
    
class wei_ang_migration(wei_ang):
  def __init__(self,pars,prog=None,**kw):
    """Initialize wave equation migration with ph gathers

      pars  - the parameters for the migration
      prog  - the program to run
      kw    - additional parameters for the parallel job

      See SEP.wei_oper.wei_operator for further parameter description

      pars must specify aw__n,aw__o,aw__d
      apx_n,apx_o,apx_d - ph axis description in x
      apy_n,apy_o,apy_d - ph axis description in y

    """
    pars.add_string("operation=migration")
    if kw.has_key("restart"): restart=kw["restart"]
    else: restart=None
    wei_ang.__init__(self,pars,prog,**kw)
    mtag,err1=self.pars.par("R")
    dtag,err2=self.pars.par("D")
    if err1==0: SEP.util.err("When doing migration must provide model (R)")
    if err2==0: SEP.util.err("When doing migration must provide data (D)")
    self.data=SEP.sep_vector.scmplx(tag=dtag)
    self.check_data(self.data)
    if self.verb: SEP.util.msg("Initializing/reading a model")
    self.image=self.create_model(mtag)
    if not restart: 
        self.image.write_file()
        self.image.zero()
    self.restart=restart
  def migrate(self):
    """Run a migration"""
    self.adjoint_op(self.image,self.data,restart=self.restart)
