import SEP.util
import SEP.par
import SEP.vector
import SEP.prog
import time,os,string,types,sys
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2005.2.16"
                                                                                



def check_space(vec1,vec2):
    """Check to see if two vectors are of the same type and size """
    if type(vec1)!=type(vec2):
      SEP.util.msg("Vectors not of the same type")
      return 1
    if vec1.size()!=vec2.size():
      SEP.util.msg("Vectors not the same size")
      return 1
    return None


class operator:
  """A generic operator class"""

  def __init__(self,name,domain,range,**kw):
    """Initialize a generic operator
    
       name    - ASCII string describing operator (required)
       domain  - The domain vector for the operator (required)
       range   - The range vector for the operator  (required)
       verb    - Whether or not to be verbose when running operator (0)
       msg     - The message to print when running (defaults to name)

    """
    self.domain=domain.clone_space()
    self.range=range.clone_space()
    self.kw=kw
    self.name=name
    self.verb=self._option("verb",0)
    self.nsteps=1
    self.msg=self._option("msg",name)

  def _option(self,par,default=None):
    """Return an optional paramter if it exists, else None"""
    if not self.kw.has_key(par):  return default
    return self.kw[par]

  def set_verb(self,verb):
    self.verb=verb

  def adjoint(self,d,r,status=None,iter=-1,add=None,restart=None):
    """Run adjoint operation

       d       -   domain vector
       r       -   range vector
       status  -   status object to update when running (optional)
       iter    -   Iteration we are on (optional=0)
       add     -   Whether or not we are adding
       restart -   Whether or not we are restarting

    """

    name="%d.%s_a"%(iter,self.name)
    if restart and name!=restart: return restart
    self.check_operator_logic(d,r)
    if status: status.update_status(name,["started"],lock=1)
    if self.verb: SEP.util.msg( "%d iter adjoint operation %s"%(iter,self.msg))
    self.adjoint_op(d,r,add,restart)
    if status: status.update_status(name,["finished"],lock=1)
    if  restart: return None
    return restart

  def forward(self,d,r,status=None,iter=-1,add=None,restart=None):
    """Run forward operation

       d       -   domain vector
       r       -   range vector
       status  -   status object to update when running (optional)
       iter    -   Iteration we are on (optional=0)
       add     -   Whether or not we are adding
       restart -   Whether or not we are restarting

    """

    self.check_operator_logic(d,r)
    name="%d.%s_f"%(iter,self.name)
    if restart and name!=restart: return restart
    if self.verb: SEP.util.msg( "%d iter forward operation %s"%(iter,self.msg))
    if status: status.update_status(name,["started"],lock=1)
    self.forward_op(d,r,add,restart)
    if status: status.update_status(name,["finished"],lock=1)
    if  restart: return None
    return restart

  def init_op(self,restart):
    """Initialize the operator"""

  def adjoint_op(self,model,data,add,restart):
    """Function to run the adjoint operation"""
    SEP.util.err("Must define the adjoint operation")

  def forward_op(self,model,data,add,restart):
    """Function to run the forward operation """
    SEP.util.err("Must define the forward operation")
    
  def job_desc(self,iter):
    """Return job description for an operator"""
    return ["%d.%s_f"%(iter,self.name)], ["%d.%s_a"%(iter,self.name)]
    
  def check_operator_logic(self,d,r):
    """Check to make sure the doman and range vector are correct for the operator"""
    if check_space(self.domain,d):
       SEP.util.err("Domain vector %s not the correct space for operator %s"%(d.name,self.name))
    if check_space(self.range,r):
       SEP.util.err("Range vector %s not the correct space for operator %s"%(d.name,self.name))


  def clean(self):
    """Perform any cleaning needed for this operation"""

  def dot_test(self):
     """Check to see if the operator passes the dot product test"""
     mod1=self.domain.clone_space(name="dot1_mod")
     mod2=self.domain.clone_space(name="dot2_mod")
     dat1=self.range.clone_space(name="dot1_dat")
     dat2=self.range.clone_space(name="dot2_dat")
     dat1.zero();dat2.zero();dat2.random()
     mod1.zero();mod2.zero();mod1.random()
     self.init_op(None)
     dot1=[];dot2=[]
     self.forward(mod1,dat1);dot1.append(dat1.dot(dat2))
     self.adjoint(mod2,dat2);dot1.append(mod1.dot(mod2))
     err=0
     if abs(dot1[0])==0.: 
       SEP.util.msg("Dot product is 0")
       err=1 
     if(abs((dot1[0]-dot1[1])/dot1[0]) >.001):
       SEP.util.msg("Failed dot product %s %s "%(dot1[0],dot1[1]));
       err=1 
     self.forward(mod1,dat1,add=1); dot2.append(dat1.dot(dat2))
     self.adjoint(mod2,dat2,add=1); dot2.append(mod1.dot(mod2))
     if abs(dot2[0])==0:
       SEP.util.msg("Dot product is 0")
       err=1 
     if(abs((dot2[0]-dot2[1])/dot2[0]) >.001):
       SEP.util.msg("Failed add dot product %s %s "%(dot2[0],dot2[1]));
       err=1 
     if not err:
       SEP.util.msg("Passed dot product!")
       return None
     else: return 1

   

  def nops(self):
    """Return the number of operators this operator is composed of"""
    return self.nsteps 

class oc(operator,SEP.prog.prog):
  """An operator class for simple out of core operations"""

  def __init__(self,prog,domain,range,**kw):
    """Initialize a simple out of core operator 
 
       prog    - the program to run
       domain  - the domain vector
       range   - the range vector
       name    - the name associated with the operator (defaults to prog)
      
       See parent classes for additional arguments

    """
    operator.__init__(self,prog,domain,range,**kw)
    self.pars=self._option("pars",SEP.par.sep_pars(name=prog))
    SEP.prog.prog.__init__(self,prog, self.name,self.pars);
    self.prog=prog
    self.check_operator_pars()


  def check_operator_pars(self):
     self.add_com=self._option("add_com","add=y")
     self.restart_com=self._option("restart_com","restart=y")
     self.adjoint_com=self._option("adjoint_com","adj=y")
     self.data_tag=self._option("data_tag","data=")
     self.model_tag=self._option("model_tag","model=")


  def forward_op(self,model,data,add,restart):
    """Run the forward operation
 
        model  - the model vector
        data   - the data  vector
        add    - whether or not we are adding
        restart- whether or not we are restarting

    """
    par=self.forward_pars(add,restart,[model.extra_args,data.extra_args])
    par.add_string("%s%s"%(self.data_tag,data.name))
    par.add_string("%s%s"%(self.model_tag,model.name))
    self.run_job(restart,par)


  def adjoint_op(self,model,data,add,restart):
    """Run the adjoint operation
 
        model  - the model vector
        data   - the data  vector
        add    - whether or not we are adding
        restart- whether or not we are restarting

    """
    par=self.adjoint_pars(add,restart,[model.extra_args,data.extra_args])
    par.add_string("%s%s"%(self.data_tag,data.name))
    par.add_string("%s%s"%(self.model_tag,model.name))
    self.run_job(restart,par)
   
  def forward_pars(self,add=None,restart=None,extra=[]):
    """Return the command to run if doing the forward operation

        add      (None) whether or not we are adding
        restart  (None) whether or not we are restarting
        extra    (None) an additional string for running command
       
    """
    par=SEP.par.sep_pars(name=self.name)
    par.add_pars(self.pars)
    if add: par.add_string(self.add_com)
    if restart: par.add_string(self.restart_com)
    if extra: 
      for e in extra: 
        if e: par.add_string(e)
    return par

  def adjoint_pars(self,add=None,restart=None,extra=None):
    """Return the command to run if doing the adjoint operation

        add      (None) whether or not we are adding
        restart  (None) whether or not we are restarting
        extra    (None) an additional string for running command
       
    """
    par=self.forward_pars(add,restart,extra)
    par.add_string(self.adjoint_com)
    return par

class oc_par(oc):
  """An operator class for simple coarse grain parallelization operators"""
                                                                                            
  def __init__(self,name,domain,range,parjob,**kw):
    """Initialize an out of core parallilzed job

        name   - the name for the operation
        domain - the domain vector for the parallel operator
        range  - the range vector for the parallel operator
        parjob - the parallel job object 
        par    - the parameters for the job (optional)

        See parent class for optional parameters

    """ 

    SEP.operator.oc.__init__(self,name,domain,range,**kw)
                                                                                            
    self.parjob=parjob
    self.parfiles=[]
    self.pars=self._option("pars",SEP.par.sep_pars(name=name))
                                                                                            
    if not parjob.return_par_files().has_key("domain"):
      SEP.util.err("Parjob must be initialized with a domain file tag")
    if not parjob.return_par_files().has_key("range"):
      SEP.util.err("Parjob must be initialized with a range file tag")
    self.range_par=parjob.return_par_files()["range"]
    self.domain_par=parjob.return_par_files()["domain"]
                                                                                            

  def forward_op(self,model,data,add=None,restart=None):
    """Run the forward operation
 
        model  - the model vector
        data   - the data  vector
        add    - whether or not we are adding
        restart- whether or not we are restarting

    """
    par=self.forward_pars(add,restart,[model.extra_args,data.extra_args])
   
    self.parjob.set_global_par_dict(self.forward_pars(
      restart=restart))

    self.parjob.update_parfile_dict("domain",
      self.domain_par.clone(name=model.name,usage="INPUT",restart=restart))

    self.parjob.update_parfile_dict("range",
      self.range_par.clone(name=data.name,usage="OUTPUT",restart=restart,add=add))

    self.parjob.run_job(restart)
    self.parjob.clean_files()
                                                                                            
  def adjoint_op(self,model,data,add=None,restart=None):
    """Run the adjoint operation
 
        model  - the model vector
        data   - the data  vector
        add    - whether or not we are adding
        restart- whether or not we are restarting

    """
    par=self.adjoint_pars(add,restart,[model.extra_args,data.extra_args])

    self.parjob.set_global_par_dict(self.adjoint_pars(
      restart=restart))

    self.parjob.update_parfile_dict("domain",
      self.domain_par.clone(name=model.name ,usage="OUTPUT",restart=restart,add=add))


    self.parjob.update_parfile_dict("range",
      self.range_par.clone(name=data.name,usage="INPUT",restart=restart))

    self.parjob.run_job(restart)
    self.parjob.clean_files()



class array(operator):
  """Complex array operator

       With this class you can build operator arrays 

        d   L 
          =    m
        0   A

  """
  def __init__(self,name,array,nrow,ncol,restart=None,verb=0,msg=None):
    """Initialize an array operator 

        name   - the name associated with the operator
        array  - a vector of operators
        ncol   - number of collumns in operator
        restart- whether or not we are restarting
        verb   - whether or not to be verbose
        msg    - Message (defaulting to name) to print when running operator

        Example:

         d   L        op=SEP.operator.array("reg",[L,A]a,1,2)
           =    m
         0   A


    """ 

    if len(array) != nrow*ncol: 
      SEP.util.err("Error: nrow=%d ncol=%d len(array)"%(nrow,ncol,len(array)))

    self.name=name
    self.op_list=array
    self.nrow=nrow
    self.ncol=ncol
    self.verb=verb
    if msg: self.msg=msg
    else: self.msg=name
    self.nop=len(array)

    self.domain_list=[]
    for irow in range(self.nrow):  self.domain_list.append(None)

    self.range_list=[]
    for icol in range(self.ncol):  self.range_list.append(None)

    i=0
    for icol in range(ncol):
      for irow in range(nrow):
        if self.op_list[i]:
          if self.range_list[icol]: 
            if check_space(self.range_list[icol],self.op_list[i].range):
              SEP.util.err("range vectors not equiv")
          else: self.range_list[icol]=self.op_list[i].range.clone_space()
          if self.domain_list[irow]: 
            if check_space(self.domain_list[irow],self.op_list[i].domain):
              SEP.util.err("domain vectors not equiv")
          else: self.domain_list[irow]=self.op_list[i].domain.clone_space()
        i=i+1

    for irow in range(nrow): 
      if not self.domain_list[irow]: SEP.util.err("Undefined domain vector %d"%irow)

    for icol in range(ncol): 
      if not self.range_list[icol]: SEP.util.err("Undefined range vector %d"%icol)
    
    if ncol >1: self.range=SEP.vector.super(self.name+".range",self.range_list)
    else: self.range=self.range_list[0].clone_space()

    if nrow >1: self.range=SEP.vector.super(self.name+".domain",self.domain_list)
    else: self.domain=self.domain_list[0].clone_space()

  def job_desc(self,iter):
    """Return job description for a matrix operator"""
    list_f=[]
    list_a=[]
    for i in range(0,self.nop,1):
      a,b=self.op_list[i].job_desc(iter)
      list_f.extend(a)
    for i in range(self.nop-1,-1,-1):
      a,b=self.op_list[i].job_desc(iter)
      list_a.extend(b)
    return list_f,list_a
    

  def forward(self,d,r,status=None,iter=0,add=None,restart=None):
    """Run forward operation for an array operator

       d       -   domain vector
       r       -   range vector
       status  -   status object to update when running (optional)
       iter    -   Iteration we are on (optional=0)
       add     -   Whether or not we are adding
       restart -   Whether or not we are restarting

    """
    if self.verb: SEP.util.msg( "Forward array operation %s"%self.msg)

    for icol in range(self.ncol):
      addit=add
      for irow in range(self.nrow):
        if self.ncol >1:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].forward(d.vecs[irow],r.vecs[icol],
             status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].forward(d,r.vecs[icol],
             status, iter,add=addit,restart=restart)
        else:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].forward(d.vecs[irow],r,
             status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].forward(d,r,
             status, iter,add=addit,restart=restart)
        addit=1
    return restart
    
  def adjoint(self,d,r,status=None,iter=0,add=None,restart=None,stat=1):
    """Run adjoint operation for an array operator

       d       -   domain vector
       r       -   range vector
       status  -   status object to update when running (optional)
       iter    -   Iteration we are on (optional=0)
       add     -   Whether or not we are adding
       restart -   Whether or not we are restarting

    """
    if self.verb: SEP.util.msg( "Adjoint array operation %s"%self.msg)
    for irow in range(self.nrow-1,-1,-1):
      addit=add
      for icol in range(self.ncol-1,-1,-1):
        if self.ncol >1:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d.vecs[irow],r.vecs[icol],
              status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d,r.vecs[icol],
              status, iter,add=addit,restart=restart)
        else:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d.vecs[irow],r,
              status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d,r,
              status, iter,add=addit,restart=restart)
        addit=1
    return restart
    
 
  def init_op(self,restart):
    """Initialize an array operator"""
    for op in self.op_list: op.init_op(restart)

  def clean(self):
    """Clean an array opeator"""
    if self.verb: SEP.util.msg("Clean matrix operation %s"%self.msg)
    for op in  self.op_list: op.clean()


class chain(operator):
  """A class for chaining together operators"""

  def __init__(self,name,op_list,restart=None,verb=0, msg=None):
    """Initialize a chain operator

       name     - the name for the operator
       op_list  - the list of operators to chain together
       restart  - whether we are restarting
       verb     - whether or not to be verbose
       msg      - the message to print when running 

       Example:

          d=LBp
         op=SEP.operator.chain("prec",[L,B])

    """
    self.name=name
    if not msg: self.msg=name
    else: self.msg=msg
    self.verb=verb

    self.op_list=op_list
    self.inter_vec=[]
    self.domain=self.op_list[0].domain.clone_space()
    self.range=self.op_list[len(self.op_list)-1].range.clone_space()
    self.nsteps=0
    self.nop=len(self.op_list)
    for op in op_list: self.nsteps+=op.nops()

    for i in range(1,len(self.op_list)):
      rlen=self.op_list[i-1].range.size()
      dlen=self.op_list[i].domain.size()
      if rlen!=dlen:
        SEP.util.err("domain and range vector mismatch for chain_op")
      self.inter_vec.append(self.op_list[i].domain.clone_space(name=name+str(i)))
      

  def job_desc(self,iter):
    """Return job description for a chain operator"""
    list_f=[]
    list_a=[]
    for i in range(0,self.nop,1):
      a,b=self.op_list[i].job_desc(iter)
      list_f.extend(a)
    for i in range(self.nop-1,-1,-1):
      a,b=self.op_list[i].job_desc(iter)
      list_a.extend(b)
    return list_f,list_a
    

  def forward(self,d,r,status=None,iter=-1,add=None,restart=None):
    """Run adjoint operation for an chain operator

       d       -   domain vector
       r       -   range vector
       status  -   status object to update when running (optional)
       iter    -   Iteration we are on (optional=0)
       add     -   Whether or not we are adding
       restart -   Whether or not we are restarting

    """
    if self.verb: SEP.util.msg("%s iter forward row operation %s"%(str(iter),self.msg))
    restart=self.op_list[0].forward(d,self.inter_vec[0],status,iter,restart=restart)
    for i in range(0,self.nop-2):
      restart=self.op_list[i+1].forward(self.inter_vec[i],self.inter_vec[i+1],status,iter,restart=restart)
    restart=self.op_list[self.nop-1].forward(self.inter_vec[self.nop-2],r,status,iter,add=add,restart=restart)
    
  def adjoint(self,d,r,status=None,iter=-1,add=None,restart=None):
    """Run adjoint operation for an chain operator

       d       -   domain vector
       r       -   range vector
       status  -   status object to update when running (optional)
       iter    -   Iteration we are on (optional=0)
       add     -   Whether or not we are adding
       restart -   Whether or not we are restarting

    """
    if self.verb: SEP.util.msg( "%s iter adjoint row operation %s"%(str(iter),self.msg))
    restart=self.op_list[self.nop-1].adjoint(self.inter_vec[len(self.inter_vec)-1],r,status,iter,restart=restart)
    for i in range(self.nop-3,-1,-1):
      restart=self.op_list[i+1].adjoint(self.inter_vec[i],self.inter_vec[i+1],status,iter,restart=restart)
    restart=self.op_list[0].adjoint(d,self.inter_vec[0],status,iter,add,restart=restart)

  def clean(self):
    """Clean a chain operator (delete intermediate space)"""
    if self.verb: SEP.util.msg( "Clean row operation %s"%self.msg)
    for op in  self.op_list: op.clean()
    for vec in self.inter_vec: 
      vec.clean()

  def init_op(self,restart):
    """Initialize a chain operator"""
    for op in self.op_list: op.init_op(restart)
    for vec in self.inter_vec: 
      if not restart: vec.zero()
      else: vec.load()


class scale(operator):
  """A simple operator that does scaling by a scalar """
  def __init__(self,name,domain,range,scale,restart=None,verb=0):
     """Initialize a scale operator

       name     - the name for the operator
       domain   - the domain vector
       range    - the range vector
       scale    - the scale
       restart  - whether we are restarting
       verb     - whether or not to be verbose
       msg      - the message to print when running 

     """
     if domain.size() != range.size():
       SEP.util.err("domain and range must be the same size for scale_op")
     self.scale=scale
     operator.__init__(self,name,domain,range,verb=verb)
     
  def adjoint_op(self,d,r,add,restart=None):
    """Run the adjoint scaling operation
 
        model  - the model vector
        data   - the data  vector
        add    - whether or not we are adding
        restart- whether or not we are restarting

    """
    if add: sc=1.
    else: sc=0.
    d.scale_addscale(sc,r,self.scale)

  def forward_op(self,d,r,add,restart=None):
    """Run the scaling forward operation
 
        model  - the model vector
        data   - the data  vector
        add    - whether or not we are adding
        restart- whether or not we are restarting

    """
    if add: sc=1.
    else: sc=0.
    r.scale_addscale(sc,d,self.scale)



