import re
import string
import pwd
import sys
import os
import time
import SEP.util
import types

""" 

  Module for parameters SEP style parameter handling

  

"""
                                                                                
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2004.2.4"

                                                                                
def extract_pars(pars,prefix):
  """Extract all the parameters begining with a given prefix

     pars    - SEP.par.sep_par object
     prefix  - prefix to extract

     Returns parameter structure with parameters with prefix (prefix removed)
     
     Example: 
       parameters in  par_in contains mig_n1=23 mig_n2=12
       par_out=extract_par(par_in,"mig_")
       parameters in par_out contains n1=23, n2=12
   
   """

  out=SEP.par.sep_pars(name=prefix)
  srS=re.compile("%s(\S+)"%prefix)
  for key,val in pars.return_dict().items():
    b=srS.search(key)
    if b: 
      out.add_param(b.group(1),val)
  return out

def dict_to_list(dict,prefix=""):
  """Convert a dictionary to key=val list, add optional prefix
   
   dict   - A dictionary object
   prefix - The prefix to add (defaults to '')

   Example:
      dict_in -  {n1:23 , b1:24, c1:'alpha' >:a.H}
      list_out= dict_to_list(dict_in,"mig_")
      list_out= ["mig_n1=23","mig_b1=24","c1='alpha'"," >a.H"]

  """

  list=[]
  for key,val in dict.items(): 
    if key!=">" and key!="<": div="="
    else: div=""
    list.append("%s%s%s%s"%(prefix,key,div,val))
  return list


def logic(val):
  """Return 0 if par i 'n',"N",'0' or None else return 1"""
  if not val: return None
  val=str(val)
  valit=1
  if val[0]=='n' or val[0]=="N" or val[0]=='0': valit=None
  return valit

class sep_pars:
  """ A class to hold SEP style parameters"""

  def __init__(self,file=None,name=None,seppar=None):
     """Initializer for SEP.par.sep_pars object

        Three initialization methods:

        From a file par=SEP.par.sep_pars(file='file.P')
        From another parameter object par=SEP.par.sep_pars(seppar=par1)
        An empty list par=SEP.par.sep_Pars(name="new")
     
     """
     self.params={}
     self.added={}
     self.iamstring={}
     self.qbase="A1b2C"
     self.name="None"
     self.mylines=[]
     self.parq1=re.compile('(\S+)="(\S+)"')
     self.parq2=re.compile("(\S+)='(\S+)'")
     self.parS=re.compile('(\S+)=(\S+)')
     if file != None: 
       self.name=file
       self._build_database(file)
     elif seppar != None:  
       if name !=None: self._init_seppar(seppar,name)
       else: 
         SEP.util.err("Must provide the name when initializing a sep_par")
     elif name !=None: self.name=name
     else:
       SEP.util.err( "Must provide name or file when creating ")
       
  def extract_del_pars(self,prefix):
    """Extract from parameter object all parameters begining with prefix

   
       Example:  object par contains mig_n1=23, n2=27, a=24
       par_out=par.extract_del_pars("mig_")
       par_out contains n1=23
       par contains n2=27,a=24

    """
     
    par_out=extract_pars(self,prefix)
    for key in self.return_dict().keys(): self.del_par("%s%s"%(prefix,key))
    return par_out

  def logic(self,par,default=None,delete=None):
    """Return the status of a logical value in parameter object

       If logical parameter is n,N, or 0 returns Null
       Default parameter (None) will be returned if parameter doesn't exist
       delelte=1 Will delete the parameter

       Examples:
          par contains alpha=1, beta=n

                                    returns          par structure
          par.logic(alpha)             1              alpha=1, beta=n
          par.logic(beta,delete=1)    None            alpha=1
          par.logic(delta,1)            1             alpha=1, beta=n, delta=1
          par.logic(delta,None,1)     None            alpha=1, beta=n

     """
          
    val,err=self.par(par,default,delete=delete)
    if err==1:  return logic(val)
    return default

  def _init_seppar(self,seppar,name):
    """Initialize a seppar object from another seppar object"""

    self.add_pars(seppar)
    self.name=name
    self.mylines = seppar.mylines

  def add_pars(self,seppar):
    """Add the parameters from another par object

       par   contains a=23,b=24,c=25
       par2  contains d=21,e=a.H

       par.add_pars(par2)

       Resulting par contains a.23,b=24,c=25,d=21,e=a.H

    """
  
    for par,val in seppar.params.items():
      self.params[par]=val
      self.added[par]=val
    for par,val in seppar.iamstring.items():
      self.iamstring[par]=val

  def _build_database(self,file):
    """Build the database from a file"""
  
    try: f = open (file)   #OPEN THE FILE
    except:
      SEP.util.err( 'Trouble opening file (%s) for reading '%file)
    lines = f.readlines()  #READ IN THE ENTIRE FILE TO LINES
    for line in lines:
      self.mylines.append(line)
      self.add_string(line)
    f.close()  

  def add_string(self,line):
    """Parse a string adding paramters


       Example:
         par contains a=23,b=25
         par.add_string("c=27 d=28")
         par out a=23,b=25,c=27,d=28
      
    """
    args=line.split()
    self._parse_arg(args,1)

  def return_dict(self):
    """Return the dictionary relating to a parameter structure"""
    return self.params
  def _parse_arg(self,args,add=None):
    """Parse an argument changing a=b to par{a}=b"""
    comment=0
    for arg in args:
      if arg[0]=="#": comment=1
      res=None
      if comment!=1:
        q=0
        res=self.parq1.search(arg)
        if  res: q=1
        else:
          res=self.parq2.search(arg)
          if res: q=1
          else:
            res=self.parS.search(arg)
      if res:
        if res.group(1)=="par":
          self._build_database(res.group(2))
        else:
          self.params[res.group(1)]=res.group(2)
          if q==1: self.iamstring[res.group(1)]=1
          if add:self.added[res.group(1)]=res.group(2)


  def par(self,p,default=None,error=None,delete=None):
    """Return a parameter value  and its number of elements

       Example:
         par contains a=27   c=14.4,16
         
        command                  return          par strucutre
        par.par('a')             27,1            a=27  c=14.4,16
        par.par('b')             None,0          a=27  c=14.4,16
        par.par('c')             [14.4,16],2     a=27  c=14.4,16
        par.par('b',3)           3,1             a=27 b=3 c=14.4,16
        par.par('b',error=1)     error
        par.par('c',delete=1)    [14.4,16],2     a=27 

    """ 

    if self.params.has_key(p):
      if self.iamstring.has_key(p): a,b=self.params[p],1
      a,b= self._not_string(self.params[p])
      if delete: self.del_par(p)
      return a,b 
    elif default != None:
      self.add_param(p,default)
      if type(default) == types.StringType and len(str(default))==0:
        return default,1
      if str(default)[0]=='"' or str(default)[0]=="'": return default,1
      return self._not_string(self.params[p])
    elif error: SEP.util.err("Request for non-existent parameter %s"%p)
    else: return None,0


  def param(self,p,default=None,error=None,delete=None):
    """Return parameter value not number of elements see SEP.par.sep_pars.par"""
    return self.par(p,default,error,delete)[0]

  def del_par(self,par):
    """Delete a paramter from the parameter list"""
    if self.params.has_key(par): del self.params[par]

  def add_dictionary(self,dict,prefix=""):
    """Add a dictionary to the agument list"""
    list=dict_to_list(dict,prefix)
    self._parse_arg(list)

  def return_dict(self):
    """Return the argument dictionary"""
    return self.params
  
  def return_par_list(self):
    """Return the argument list

       Example:
         par is a=23 b=12.,13.
         par.return_par_list() returns [a=23.,b=[12.,13.]]

    """
    return dict_to_list(self.params)

  def return_basic_pars(self,par_list=None):
    """Return the argument list of basic types (string,int, float)

       Example:
         par is a=23 b=12.,13.
         par.return_sep_par_list() returns a:23.,b:"12.,13."]

    """
    if not par_list: par_list=self.params.keys()
    pstruct=SEP.par.sep_pars(name="temp")
    for p in par_list:
       val=self.params[p]
       if type(val)==types.IntType or type(val)==types.FloatType or type(val)==types.StringType or type(val)==types.ListType:
         if val: pstruct.add_param(p,val)
    return pstruct

  def return_sep_par_list(self,quote=1):
    """Return the argument list with None converting "n"

       Example:
         par is a=23 b=12.,13.
         par.return_sep_par_list() returns [a=23.,b=[12.,13.]]

    """
    alist=[]
    if quote: 
      beg="'"
      end="'"
    else:
      beg=""
      end=""
    for key,val in self.params.items():
      if type(val)!=types.NoneType:
         if key == "<" or key == ">": alist.append("%s%s"%(key,val))
         else:
           if type(val)==types.ListType:
             v=val[0]
             if type(v)==types.FloatType or type(v)==types.StringType or type(v)==types.IntType:
               b=[]
               for v in val: b.append(str(v))
               alist.append("%s%s=%s%s"%(beg,key,string.join(b,","),end))
           else: alist.append("%s%s=%s%s"%(beg,key,val,end))
    return alist

  def return_sep_par_string(self):
    """Return all of the arguments in a single string converting None to n"""
    return string.join(self.return_sep_par_list())
  
  def return_par_string(self):
    """Return the argument dictionary

       Example:
         par is a=23 b=12.,13.
         par.return_par_list() returns [a:23.,b:[12.,13.]]

    """
    return string.join(dict_to_list(self.params))
  

  def write_pars(self,file):
    """ Write the parameters to a file"""
    try: f = open (file,"w")   #OPEN THE FILE
    except:
      SEP.util.err( 'Trouble opening file for writing:', file)
    for line in self.mylines: f.write(line)
    for line in self._new_lines(): f.write(line)

  def  _new_lines(self):
    """Return parameters that have been modified or added"""
    lines=[] 
    lines.append("%s :   %s@%s  %s\n"%(sys.argv[0],pwd.getpwuid(os.getuid())[0],
      os.uname()[1],time.strftime("%a %b %d %H:%M:%S %Y",time.localtime(int(time.time())))))
    for par,item in self.added.items():
      if self.iamstring.has_key(par): lines.append("\t%s='%s'\n"%(par,str(item)))
      else: lines.append("\t"+par+'='+str(item)+"\n")
    return lines

  def default_val(self,par,val):
    """Set a default value for a parameter"""
    self.param(par,val)

  def add_param(self,par,val,type="str"):
    """Add a parameter to the parameter class
  
    Add a parameter par to the class, give it the value val.
    Optional argument for the type

    """
    if re.search("\s+",str(val)): self.iamstring[par]=1
    if type == "int":
      self.params[par]=int(val)
      self.added[par]=int(val)
    else:
      self.params[par]=val
      self.added[par]=val
     
  def _not_string(self,cc):
     """Return a non-string parameter, breaking it into a list of called for"""
     if types.ListType == type(cc): return cc,len(cc)
     if re.search(",",str(cc)):
       vals=cc.split(",")
       return vals,len(vals)
     else:
       return cc,1

class sep_args(sep_pars):
  """A SEP.par.sep_par class that is initialzed with the command line arguments"""
  def __init__(self):
    """Initialize a sep parameter object with command line arguments """
    sep_pars.__init__(self,name="args")
    self._parse_arg(sys.argv[1:])
