import SEP.sep_socket
import SEP.sepfile
import SEP.mpi
import os,sys,types,commands,string,re
import SEP.util
import SEP.datapath
import SEP.paths
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2004.12.2"
                                                                                


 
class parfile(SEP.sepfile.sep_file,SEP.status.sep_status):
  """ Define a sep parallel file"""
  def __init__(self,**kw):
     """Initialize a parfile object

        Required parameters:
        logdir      -   The directory we are logging in
        name        -   The name of the combined file
        tag         -   The tag associated with the file (e.g. data=)
        isect_dict  -   A dictionary mapping key:isect 
        usage       -   Usage for file (INPUT or OUTPUT)

        Optional parameters:
        restart     -   WHether or not we are restarting (None)
        verb        -   Verbosity level (None)
        add         -   Whether or not we are adding to the output (None)
        reuse_par   -   Parameter when a file is being reused (defaults to none)
        remove      -   Whether or not use remove the part files (defaults to yes)
        space_only  -   Whether or not we are just creating the space (defaults to no)
        distributed -   Whether or not (default) the file is already distributed in some manner
        collect_it  -   Whether or not collect the file
   
     """
 
     self.type=None
     self.kw=kw
     self.name=self._option("name",err=1)  
     self.tag=self._option("tag",err=1)
     self.isect_dict=self._option("isect_dict",err=1)
     self.usage=self._option("usage")

     self.stat_file=self._option("stat_file")
     self.add=self._option("add",None)
     self.verb=self._option("verb")
     self.reuse_par=self._option("reuse_par")
     self.space_only=self._option("space_only")
     self.remove=self._option("remove",default=1)
     self.collect_it=self._option("collect_it",default=1)
     logdir=self._option("logdir")
     if logdir: self.stat_file="%s/%s.stat"%(logdir,os.path.basename(self.name))
     else: self.stat_file=self.name+".stat"

     if self._option("distributed",default=None):
       aa=SEP.status.sep_status(self.stat_file,load=1)
       nb=aa.file_param("nblock")
       ax=aa.file_param("dff_axis")
       self.par_file=self.clone(distributed=None,dff_axis=ax,nblock=nb,restart=1)
       self.par_file.restart(1)
     else: self.par_file=None
     self.created     =None
     self.mach_func   =None
     restart=self._option("restart",default=None)


     if not self.space_only:
       self._check_run_param()
       if not self.usage: SEP.util.err("Must provide usage")
       if self.par_file:
         self.name="%s.input"%self.name
         self.stat_file="%s.input"%self.stat_file

       SEP.status.sep_status.__init__(self,self.stat_file,
         self.isect_dict.keys(),njob_extra=2)

       if not self.par_file:
         if self.usage=="INPUT" or self.add:
           SEP.sepfile.sep_file.__init__(self,file=self.name)
           self.created=1
       if self.verb>2: SEP.util.msg("    Finished initializing %s"%self.name)

  def _section_name(self,key,mlabel):
    """Return the section name given the key and machine label (must override)"""
    SEP.util.err("Section name must be overriden")

  def clone(self,**kw):
    """Clone a parallel file object (must override)"""
    SEP.util.err("Clone must be overriden")


  def set_mpi_mach(self,mpi,mach):
     """Set the machine class"""
     mach,  #machine functions
     self.mach_func        =mach
     self.mpi=mpi
     if self.par_file: self.par_file.set_mpi_mach(mpi,mach)

  def mach(self):
    """Return the machine class associated with the parallel file"""
    if not self.mach_func: SEP.util.err("Machine function not set for parfile")
    return self.mach_func

  def record(self,key):
     """Record that a job(key) has started correctly"""
     self.set_key_status_elem(key,0,"running")

  def done(self,key):
     """Mark that a section (key) has been processed"""
     self.set_key_status_elem(key,0,"finished")

  def redo(self,key,redo=None):
     """Mark that a section (key)  needs to be redone"""
     self.update_status(key,["todo","none","none"])

  def nsect(self):
    """Return the number of sections"""
    return len(self.isect_dict.keys())
    
  def _check_run_param(self):
    """Check to make sure the input parameters make sense"""
    if self.usage !="INPUT" and self.usage!="OUTPUT":
      SEP.util.err( "Distribution type must be INPUT or OUTPUT usage=%s"%self.usage)
    
  def tags(self,key_label):
    """Get local section names

        key_label is a dictionry linking key names and machine labels
        Return is a dicitonary linking key names and tags

        If the file is  "INPUT" the data will be distributed to the nodes if needed
        If the file is being reused or restarted appropriate parameters will be added

    """
    err,sect_names,new_list=self._get_section_names(key_label)
    if err: return err,None
    tag_list={}
    for key in sect_names.keys():
      tag_list[key]=self.tag+sect_names[key]
      if self.usage =="OUTPUT":
        tag_list[key]+="  datapath=%s"%self._datapath(self.mach().mach_from_label(key_label[key]))
        if self.reuse_par:
          if 0== new_list.keys().count(key):
            tag_list[key]+=" "+self.reuse_par
    return None,tag_list

  def _datapath(self,mach):
    """Return the datapath for the output file"""
    return SEP.datapath.datapath(mach)

  def _get_section_names(self,key_label):
    """Get the section names given the key list
        
       Given key:machine label returns the section names

    """
    new_mach={}
    new_sect={}
    section_list={}
    for key,mlabel in key_label.items():
      a,b=self._section_name(key,mlabel)
      section_list[key]=a
      if b: 
       new_mach[key]=self.mach().mach_from_label(mlabel)
       new_sect[key]=a
      self.record_section(key,mlabel,a)
      if self.verb >4: SEP.util.msg("tag=%s '%s' isect=%d section=%s machine=%s"%(self.name,key,self.isect_dict[key],a,mlabel))
    if new_mach.keys():
       self._remove_old_sects(new_mach,new_sect)
       if self.usage=="INPUT": 
         if self._partition(new_mach,new_sect): return 1,None,None
    return None,section_list,new_mach
   
  def _remove_old_sects(self,key_mach,key_sect):
    """Remove old sections 

        Given key:mach and key:section removes old versions of the files

    """
    for key in key_mach.keys():
       SEP.spawn.run_wait("rsh -n %s 'rm -f %s'"%(key_mach[key],key_sect[key]),
        60, error=None,quiet=1)
       #not dealing with repiped
       SEP.spawn.run_wait("rsh -n %s 'rm -f %s@'"%(key_mach[key],key_sect[key]),
        60, error=None,quiet=1)
#       SEP.spawn.run_wait("rsh -n %s 'rm -f %s%s*'"%(key_mach[key],
#        self._datapath(key_mach[key]),key_sect[key]),120, error=None,quiet=1)

  def _keys_from_machine(self,mach):
    """Return all of the sections from a given machine"""
    label_list=self.mach().label_list(mach)
    sect_on_mach=[]
    for label in label_list: sect_on_mach.extend(self.get_elem_val(1,label))
    return sect_on_mach
        

  def record_section(self,key,label,section):
     """Record to the status file that we have set the machinename"""
     self.update_status(key,["sent",label,section,"none"])


  def finish(self):
    """Finish processing a file"""
    if self.usage=="OUTPUT" and self.collect_it: 
      self.collect(self.isect_dict.keys())
    
  def _option(self,par,default=None,kw=None,err=None):
    """Return an optional paramter if it exists, else None"""
    if kw==None: kw=self.kw
    if not kw.has_key(par):
      if err: SEP.util.err("Missing required parameter %s"%par)
      return default
    return self.kw[par]

  def _partition_pars(self,key_mach,key_sect):
    """Return the command line parameters for distributing a dataset

       key_mach a dictionary key:machine
       key_sect a dictionary key:section name

       returns a list that is command line parameters to partition the dataset

    """
    SEP.util.err("Must override partition pars")

  def _collect_pars(self,key_mach,key_sect):
    """Return the command line parameters for collecting a dataset

       key_mach a dictionary key:machine
       key_sect a dictionary key:section name

       returns a list that is command line parameters to collect the dataset

    """
    SEP.util.err("Must override collect pars")

  def _partition(self,key_mach,key_sect):
     """Partition the dataset"""
     if self.usage != "INPUT": SEP.util.err("Can only partition INPUT")
     pars,machs,nbytes=self._partition_pars(key_mach,key_sect)
     if self.verb >1: SEP.util.msg("Distributing %s to %s"%(self.name,
         string.join(key_mach.values())))
     return self.mpi.run_job(pars,machs,nbytes)

  def collect(self,key_list,extra_args=[" "]):
     """Collect the dataset"""
     if self.usage != "OUTPUT" or not self.collect_it: return
     pars,machs,nbytes=self._collect_pars(key_list)
     if self.verb >1: SEP.util.msg("Collecting  %s from %s"%(self.name,
        string.join(machs)))
     return self.mpi.run_job(pars,machs,nbytes)

  def _remove_parts(self):
    """Remove the sections"""
    if not self.remove: return
    mach_rm={}
    for key in self.return_keys():
      mach= self.mach().mach_from_label(self.get_key_status_elem(key,1))
      if not mach_rm.has_key(mach): mach_rm[mach]=[]
      tag= self.get_key_status_elem(key,2)
      tfile=SEP.sepfile.sep_file(remote="%s:%s"%(mach,tag))
      mach_rm[mach].append(tag)           #remove the history file
      mach_rm[mach].append(tfile.history.param("in"))  #remove the data binary
      if tfile.headers:
        mach_rm[mach].append(tfile.history.param("hff"))  #remove the hff
        mach_rm[mach].append(tfile.headers.param("in"))  #remove the headers binary
      if tfile.grid:
        mach_rm[mach].append(tfile.history.param("gff"))  #remove the grid file
        mach_rm[mach].append(tfile.grid.param("in"))  #remove the grid binary

    icount=0
    for mach,tags in mach_rm.items():
      icount+=1 
      if icount==6:
        icount=0; 
        SEP.spawn.run_wait("rsh -n %s 'rm -f %s '"%(mach,string.join(tags)));
      else:
        SEP.spawn.run_wait("rsh -n %s 'rm -f %s &'"%(mach,string.join(tags)));
        

  def clean(self):
    """Clean all the intermediate portions"""
    if self.par_file: self.par_file.clean()
    if self.remove: 
      self._remove_parts()
      SEP.status.sep_status.clean(self)
    dirs=[".",os.environ.get('HOME')]
    ab=re.compile("^[\.\w _\-@:~]+$")
    for dir in dirs:
      for file in os.listdir(dir):
        filep=dir+"/%s"%file
        st=os.stat(filep)
        if st.st_size==0 and not ab.search(file):
          os.unlink(filep)

class distribute(parfile):
  """A parallel file that is broken up along one or more axis"""

  def __init__(self,**kw):
    """Initialze a distributed file
    
       Required parameters:

       dff_axis - the axis/axes that the data is distributed along
       nblock   - the number of sections along each the dataset is broken into

       Optional parameters:
       partition_prog - the program used to partion the data (Patch_split)
       collect_prog - the program used to partion the data (Patch_join)

       SEE parent class for more parameters

    """

    parfile.__init__(self,**kw)
    self.type="DISTRIBUTE"
    dff_axis=self._option("dff_axis",err=1)  
    if type(dff_axis)!=types.ListType: dff_axis=[dff_axis]
    self.dff_axis=[]
    for i in dff_axis: self.dff_axis.append(str(i))
    nblock=self._option("nblock",err=1)  
    if type(nblock)!=types.ListType: nblock=[nblock]
    self.add_file_param("nblock",nblock)
    self.add_file_param("dff_axis",dff_axis)
    self.noverlap=[]
    self.nblock=[]
    for i in nblock: 
      self.nblock.append(str(i))
      self.noverlap.append(str(0))
    if len(self.nblock) != len(self.dff_axis):
      SEP.util.err("Length of nblock and dff_axis not the same for %s"%self.name)
    part="%s/Patch_split"%SEP.paths.sepbindir
    combo="%s/Patch_join"%SEP.paths.sepbindir
    self.partition_prog=self._option("partition_prog",part)
    self.collect_prog=self._option("collect_prog",combo)

  def clone(self,**kw):
    """Clone a distributed file object"""
    kout={}
    for a,b in self.kw.items(): kout[a]=b
    for key,val in kw.items(): kout[key]=val
    if not kw.has_key("space_only") and kout.has_key("space_only"):  
      del kout["space_only"]
    a=distribute(**kout)
    return a

  def _section_name(self,key,mlabel):
     """Get the file name if we have distributed a file

        key   - the section name
        mlabel- the machine label

        returns section_name,section_name if we have a new section
        returns section_name,None if the section alread exists


     """
     name= self.get_key_status_elem(key,2)
     if name and name!="none": 
       machl=self.get_key_status_elem(key,1)
       if machl and machl!="none":
         mach_old=self.mach().mach_from_label(self.get_key_status_elem(key,1))
         mach_new=self.mach().mach_from_label(mlabel)
         if mach_old == mach_new: return name,None
     name=self._form_sect_name(key,self.mach().mach_from_label(mlabel))
     return name,name 

  def _form_sect_name(self,key,mach):
     """Get the section name given the key and machinename"""
     return "%s%s_par_DFF_%s"%(self._datapath(mach),os.path.basename(self.name),
     str(self.isect_dict[key]))

  def _distributed_pars(self,pars):
    """Add parameters if the file is distributed"""
    if self.par_file: 
      pars,machs,mach_ithread=self.par_file.add_pars_all_sections(self.return_keys(),pars)
    else:
      mach_ithread={}; machs=[] 
    pars.append("input_begin=%d"%(len(machs)+1))
    return pars,mach_ithread,machs
      


  def _partition_pars(self,key_mach,key_sect):
    """Return the command line parameters for distributing a dataset

       key_mach a dictionary key:machine
       key_sect a dictionary key:section name

       returns a list that is command line parameters to partition the dataset

    """
    pars=[self.partition_prog,"combo=%s"%self.name]
    pars,mach_distrib,machs=self._distributed_pars(pars)
    pars.append("out_dff_axis=%s"%string.join(self.dff_axis,","))
    pars.append("out_noverlap=%s"%string.join(self.noverlap,","))
    pars.append("out_nblock=%s"%string.join(self.nblock,","))
    i=0;
    imach=len(machs)
    mach_thread={}
    for key,mach in key_mach.items():
      pars.append("out_tag_sect%d=%s"%(i,key_sect[key]))
      pars.append("out_datapath_sect%d=%s"%(i,self._datapath(mach)))
      if not mach_thread.has_key(mach):
        imach=imach+1;
        mach_thread[mach]=imach;
        machs.append(key_mach[key])
      pars.append("out_sect_thread%d=%d"%(i,mach_thread[mach]))
      pars.append("out_isect_sect%d=%s"%(i,self.isect_dict[key]))
      i=i+1
    if self.par_file:
      key=self.par_file.return_keys()[0]
      sect=self.par_file.get_key_status_elem(key,2)
      mach=self.par_file.mach().mach_from_label(self.get_key_status_elem(key,1))
      first=SEP.sepfile.sep_file(remote="%s:%s"%(mach,sect))
      nbytes=first.size()*len(self.return_keys())*8
    else: nbytes=i*self.size()*8
    pars.append("nsect=%d"%i)
    pars.append("out_nsect=%d"%i)
    for i in self.nblock: nbytes=nbytes/int(i)
    return pars,machs,nbytes

  def _collect_pars(self,keys):
    """Return the command line parameters for collecting a dataset

       key_mach a dictionary key:machine
       key_sect a dictionary key:section name

       returns a list that is command line parameters to collect the dataset

    """
    pars=[]
    pars.extend([self.collect_prog,"combo=%s"%self.name])
    pars,machs,mach_ithread=self.add_pars_all_sections(keys,pars)
    if self.add: pars.append("add=1")
    key=self.return_keys()[0]
    sect=self.get_key_status_elem(key,2)
    mach=self.mach().mach_from_label(self.get_key_status_elem(key,1))
    first=SEP.sepfile.sep_file(remote="%s:%s"%(mach,sect))
    nbytes=first.size()*len(self.return_keys())*8
    return pars,machs,nbytes



  def add_pars_all_sections(self,keys,pars):
    """Return the parameters for the entire dataset
       Useful when combining a dataset or using a distributed dataset """
    pars.append("in_dff_axis=%s"%string.join(self.dff_axis,","))
    pars.append("in_nblock=%s"%string.join(self.nblock,","))
    pars.append("in_noverlap=%s"%string.join(self.noverlap,","))
    mach_ithread={}
    machs=[]
    nsect=0
    nmach=0
    for  key in keys:
      if "none"==self.get_key_status_elem(key,1):
        SEP.util.err("Internal error: collecting=%s key=%s "%(self.name,key))
      mach=self.mach().mach_from_label(self.get_key_status_elem(key,1))
      sect=self.get_key_status_elem(key,2)
      if not mach_ithread.has_key(mach):
        nmach=nmach+1
        mach_ithread[mach]=nmach
        machs.append(mach)
      pars.append("in_tag_sect%d=%s"%(nsect,sect))
      pars.append("in_sect_thread%d=%s"%(nsect,mach_ithread[mach]))
      pars.append("in_isect_sect%d=%s"%(nsect,self.isect_dict[key]))
      nsect=nsect+1
    pars.append("in_nsect=%d"%nsect)
    return pars,machs,mach_ithread


     


#PATCHING CLASS
class patch(distribute):
  def __init__(self,**kw):
    """Initialze a patched file
    
       Required parameters:

       noverlap  - the overlap 

       Optional parameters:
       partition_prog - the program used to partion the data (Patch_split)
       collect_prog - the program used to partion the data (Patch_join)

       SEE parent class for more parameters

    """
    distribute.__init__(self,**kw)
    self.type="PATCH"

    noverlap=self._option("noverlap",err=None)  
    if not noverlap:
      noverlap=[]
      for i in range(len(self.nblock)): noverlap.append(0)
    if type(noverlap)!=types.ListType: self.noverlap=[self.noverlap]
    self.noverlap=[]
    for i in noverlap: self.noverlap.append(str(i))
    if len(self.nblock) != len(self.noverlap):
      SEP.util.err("Length of nblock and noverlap not the same for %s"%self.name)
    self.partition_prog=self._option("partition_prog","%s/Patch_split"%SEP.paths.sepbindir)
    self.collect_prog=self._option("collect_prog","%s/Patch_join"%SEP.paths.sepbindir)

  def clone(self,**kw):
    """Clone a parallel file patch object"""
    kout=self.kw
    for key,val in kw.items(): kout[key]=val
    if not kw.has_key("space_only") and kout.has_key("space_only"):  
      del kout["space_only"]
    a=patch(**kout)
    return a

#SEP3D CLASS
class sep_irreg(patch):
  def __init__(self,**kw):
    """Initialze a SEP3d Irregular file
    
       Required parameters:


       Optional parameters:
       noverlap  - the overlap (defaults to 0)
       partition_prog - the program used to partion the data (Sep3d_split)
       collect_prog - the program used to partion the data (Distribute_join)

       SEE parent class for more parameters

    """
    distribute.__init__(self,**kw)
    self.type="IRREG"

    noverlap=self._option("noverlap","XXX")  
    if noverlap=="XXX": 
       noverlap=[]
       for i in self.nblock: noverlap.append(0)
    if type(noverlap)!=types.ListType: noverlap=[noverlap]
    self.noverlap=[]
    ic=0;
    for i in noverlap: 
      self.noverlap.append(str(i))
      ic+=i 
    if len(self.nblock) != len(self.noverlap):
      SEP.util.err("Length of nblock and noverlap not the same for %s"%self.name)
    self.partition_prog=self._option("partition_prog","%s/Sep3d_split"%SEP.paths.sepbindir)
    if self.usage != "INPUT" and ic>0:
      SEP.util.err("Can't  have an irregular file, patch it, and use if or output")
    self.collect_prog=self._option("collect_prog","%s/Distribute_join"%SEP.paths.sepbindir)

  def clone(self,**kw):
    """Clone a parallel file irregular sep3d object"""
    kout=self.kw
    for key,val in kw.items(): kout[key]=val
    if not kw.has_key("space_only") and kout.has_key("space_only"):  
      del kout["space_only"]
    a=sep_irreg(**kout)
    return a



class copy(parfile):
  def __init__(self,**kw):
    """Initialze a copy paralel file (same space on all nodes)
    

       Optional parameters:
       partition_prog - the program used to partion the data (Copy_split)
       collect_prog - the program used to partion the data (Copy_join)

       SEE parent class for more parameters

    """

    parfile.__init__(self,**kw)
    self.type="COPY"
    self.isect=-1;
    self.partition_prog=self._option("partition_prog","%s/Copy_split"%SEP.paths.sepbindir)
    self.collect_prog=self._option("collect_prog","%s/Copy_join"%SEP.paths.sepbindir)

  def clone(self,**kw):
    """Clone a parallel file object (copy)"""
    kout=self.kw
    for key,val in kw.items(): kout[key]=val
    if not kw.has_key("space_only") and kout.has_key("space_only"):  
      del kout["space_only"]
    a=copy(**kout)
    return a

  def _section_name(self,key,mlabel):
     """Get the file name if we have distributed a file

        key   - the section name
        mlabel- the machine label

        returns section_name,section_name if we have a new section
        returns section_name,None if the section alread exists


     """
     name=self._form_sect_name(key,self.mach().mach_from_label(mlabel))
     for file in self.get_stat_elem(2).values():
       if file==name: return name,None
     return name,name

  def _form_sect_name(self,key,mach):
     """Get the section name given the key and machinename"""
     self.isect+=1
     return "%s%s_par_DFF_%s"%(self._datapath(mach),os.path.basename(self.name),mach)

  def _collect_pars(self,keys):
    """Return the command line parameters for collecting a dataset

       key_mach a dictionary key:machine
       key_sect a dictionary key:section name

       returns a list that is command line parameters to collect the dataset

    """
    """Return the command line arguments when sharing a file"""
    pars=[self.collect_prog,"combo=%s"%self.name]
    if self.add: pars.append("add=1")
    mach_ithread={}
    sect_list=[]
    mach_list=[]
    nsect=0
    nmach=0
    machs=[]
    for  key in keys:
      mach=self.mach().mach_from_label(self.get_key_status_elem(key,1))
      sect=self.get_key_status_elem(key,2)
      if not mach_ithread.has_key(mach):
        nmach=nmach+1
        mach_ithread[mach]=nmach
        machs.append(mach)
        pars.append("in_tag_sect%d=%s"%(nsect,sect))
        pars.append("in_sect_thread%d=%s"%(nsect,mach_ithread[mach]))
        nsect=nsect+1
    #grab the first section 
    pars.append("in_nsect=%d"%(nsect))
    first=SEP.sepfile.sep_file(remote="%s:%s"%(mach,sect))
    nbytes=first.size()*nmach*8
    return pars,machs,nbytes

  def _partition_pars(self,key_mach,key_sect):
    """Return the command line parameters for distributing a dataset

       key_mach a dictionary key:machine
       key_sect a dictionary key:section name

       returns a list that is command line parameters to partition the dataset

    """
    pars=[self.partition_prog,"combo=%s"%self.name]
    i=0;
    machs=[] 
    for key,mach in key_mach.items():
      pars.append("out_tag_sect%d=%s"%(i,key_sect[key]))
      machs.append(key_mach[key])
      pars.append("out_datapath_sect%d=%s"%(i,self._datapath(mach)))
      i=i+1
    nbytes=i*self.size()*8
    return pars,machs,nbytes

class copy_share(copy):

  def __init__(self,**kw):
    """Initialze a copy paralel file (same space on all nodes) that
        is stored on a nfs disk
    

       Optional parameters:
       share_datapath -  the datapath that the image file are store on (datapath)
       nmax_procs     -  the maximum number of processes that share a file
       extra_args     -  additional arguments that need to be added with this dataset

       SEE parent class for more parameters

    """
    copy.__init__(self,**kw)
    self.dpath=self._option("share_datapath",SEP.datapath.datapath())
    self.datafile=self.dpath+os.path.basename(self.name)
    self.nmax=int(self._option("nmax_procs",8)  )
    self.big_args=self._option("extra_args","")
    self.nshare=[]
    for i in range(10000/self.nmax): self.nshare.append(0)

  def clone(self,**kw):
    """Clone a parallel file object of copy share"""
    kout=self.kw
    for key,val in kw.items(): kout[key]=val
    if not kw.has_key("space_only") and kout.has_key("space_only"):  
      del kout["space_only"]
    a=copy_share(**kout)
    return a

  def _datapath(self,mach):
    """Reeturn the datapath for the  output file"""
    return self.dpath

  def _section_name(self,key,mlabel):
    """Get the file name if we have distributed a file

        key   - the section name
        mlabel- the machine label

        returns section_name,section_name if we have a new section
        returns section_name,None if the section alread exists


    """
    for i in range(10000/self.nmax): 
      file="%s_%d"%(self.datafile,i)
      if self.nshare[i] <self.nmax:
         self.nshare[i]=self.nshare[i]+1
         if self.get_elem_val(2,file):
           return file,None
         if self.nshare[i] ==1: return file,file
         return file,None

  def done(self,key):
    """Mark that a key is finished"""
    parfile.done(self,key)
    sect=self.get_key_status_elem(key,2)
    ab=re.compile("%s_(.+)"%self.datafile).search(sect)
    if not ab:
      SEP.util.error("Trouble finding %s in %s "%(self.datafile,sect))
    self.nshare[int(ab.group(1))]=self.nshare[int(ab.group(1))]-1

  def collect(self,key_list,extra_args=[" "]):
     """Collect a dataset

          key_list is the list of keys to collect

     """
     if self.usage != "OUTPUT": return
     pars,machs,nbytes=self._collect_pars(key_list)
     if self.verb >1: SEP.util.msg("Collecting  %s from %s"%(self.name,
        string.join(machs)))
     sects=[]
     for key in key_list:
       sect=self.get_key_status_elem(key,2)
       if sects.count(sect)==0: sects.append(sect)
     t=SEP.util.temp_file()
     SEP.spawn.run_wait("Add %s >%s %s"%(string.join(sects),self.name,
       self.big_args),logfile=t) 
     SEP.spawn.run_wait("rm -f  %s %s_* &"%(t,self.datafile),60)

  def tags(self,key_label):
    """Get local section names

        key_label is a dictionry linking key names and machine labels
        Return is a dicitonary linking key names and tags

        If the file is  "INPUT" the data will be distributed to the nodes if needed
        If the file is being reused or restarted appropriate parameters will be added

    """
    err,tags=parfile.tags(self,key_label) 
    tout={}
    if not err and tags:
      for key,tag in tags.items(): tout[key]="%s %s"%(tag,self.big_args)
    return err,tout

