import SEP.sep_socket
import SEP.status
import SEP.par_log
import SEP.par
import SEP.machine
import SEP.datapath
import SEP,mpi
import SEP.par_prog
import SEP.options
import SEP.prog
import os,sys,string,time,commands,signal,types,re,pwd
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2005.02.09"
                                                                                


def simple_section_list(nblock):
  """Create a dictionary linking section descriptors and section numbers"""
  isect_dict={}
  for i in range(int(nblock)): isect_dict[str(i)]=i
  return isect_dict
  

def simple_parameter_list(nblock):
  """Create a dictionary linking section descriptors and parameter objects"""
  sect_pars={}
  for i in range(int(nblock)):
    sect_pars[str(i)]=SEP.par.sep_pars(name=str(i))
  return sect_pars


def decode_action(msg):
   """Decode  actions"""
   parts=msg.split(":")
   if len(parts)==4: return parts[0],parts[1],parts[2],parts[3]
   else: return "none","none","none","none"

def encode_action(part1,part2,part3="none",part4="none"):
   """Encode the actions returned by the server"""
   return string.join([part1,part2,part3,part4],":")


class msg_obj:
  """A Class for storing messages"""
  def __init__(self,logfile):
    """Initialzize a message object 
    
       logfile - the file to store messages from the server

    """
    self.file=logfile
    self.msg_desc =SEP.util.temp_desc("msgs")
    self.started=1

  def clean(self):
    """Remove the message file"""
    SEP.spawn.run_wait("rm %s"%self.file,50)

  def init(self):
    """Initialize the message file"""
    SEP.spawn.run_wait("rm %s;touch %s"%(self.file,self.file),50)

  def read_delete(self):
     """Read the message file contents,clear the file, and return the messages"""
     SEP.util.lock_desc(self.msg_desc)
     lines,err=self.read_msgs()
     self.write_msgs()
     SEP.util.unlock_desc(self.msg_desc)
     return lines,err

  def read_msgs(self):
    """Read the messages"""
    try: f=open(self.file)
    except: SEP.util.err("trouble reading msg file %s"%self.file)
    lines=f.readlines()
    f.close()
    return lines,None

  def write_msgs(self,lines=None):
    """Write the message file"""
    try: f=open(self.file,"w")
    except: SEP.util.err("trouble opening msg file %s"%self.file)
    if lines:
      for line in lines: f.write("%s"%line)
    f.close()


  def action(self,msg):
     """Perform an action based on a message (msg)"""
     oper,jobid,machid,extra=decode_action(msg)
     if  oper=="exit": 
       self.started=None
       return None
     SEP.util.lock_desc(self.msg_desc)
     lines,err=self.read_msgs()
     lines.append("%s\n"%msg)
     self.write_msgs(lines)
     SEP.util.unlock_desc(self.msg_desc)
     if err: return None
     return 1

class server_msg_obj(msg_obj,SEP.sep_socket.sep_server,SEP.options.par_group):
  """A Class for storing messages and communicating with processes"""
  def __init__(self,logfile,nmax,**kw):
     """A class that handles messages from a socket connection

        logfile - the file where the message stored
        nmax    - the maximum number of processes that can share a socket

        SEE parent classes for more options

     """
     msg_obj.__init__(self,logfile)
     self.nmax=nmax
     self.label_list=[]
     SEP.sep_socket.sep_server.__init__(self,**kw)
     SEP.options.par_group(name="PARJOB")
     self.started=None

  def remove_use(self,mlabel):
    """Mark that the given machine label is no longer using this socket"""
    if self.label_list.count(mlabel)==1:
      del self.label_list[self.label_list.index(mlabel)]
    else: SEP.util.error("Internal error, deleting non-used mlabel")

  def add_use(self,mlabel,port_num):
    """Try to add the given process to this socket

       mlabel   -  the machine label to attempt to add
       port_num -  the port number to attempt to use 
       
       Returns stat, pid, next_port 
        
       stat     - 1 if it can be added otherwise None
       pid      - the process id for the forked process
       next_port- next free port

    """
    if len(self.label_list) < self.nmax:
      self.label_list.append(mlabel)
      if not self.started:
         self.started=1
         pid,next_port=self.run_server(port_num)
         if not pid: sys.exit(0)
         return 1,pid,next_port
      return  1,None,None
    return None,None,None






class parjob(SEP.status.sep_status,SEP.par_log.jobs):
  """A parallel object"""

  def __init__(self,  **kw):
    """Initialize a parallel job object

       Required parameters:
       files -  a dictionary of a parallel files
       sect_pars   - the parameters associated with the different jobs
       program     - the  program to run in parallel 
 
       Optional parameters:
         
       nmax_talk - the maximum number of processes for each sect (60)
       port      - the first portion to look for an open connection (50000)
       redo      - whether or not to redo jobs that finished on node if the node dies
       status_interval- sleep time in job loop (between checking message files) (.23)
       alive_interval-how often to check if the nodes are working correctling (600)
       verb        - how verbose to be, default 0
       logdir      - the directory we send logging messages into (stat)
       shell       - the shell command to use ('rsh -n')  
       restart_com - the command to use if we are restarting a job ("")
       njobs       - the maximum number of jobs to un at one time (2000)
       attempts    - the number of attempts to run a job (3)
       global_pars - parameters that every job must have (defaults to none)

    """
   

    #save the init parameters
    self.kw=kw
    self._init_check()

  def _init_check(self):
    """Check to make sure parameters make sense"""

    #the list of parallel files
    self.files=self._option("files",err=1)

    #verbosity level
    self.verb=self._option("pverb",0)
    if not self.verb: self.verb=0
    if str(self.verb)[0]=='n' or str(self.verb)[0]=='N': self.verb=0
    self.verb=int(self.verb)

    #directory to store informat
    self.logdir=self._option("logdir","stat")
    SEP.par_log.jobs.__init__(self,self.logdir,self.verb)


    self.status_interval=float(self._option("status_interval",10.))
    self.alive_interval=float(self._option("alive_interval",900.))


    #put all of the job commands to a file
    self.all_coms="%s/%s"%(self.logdir,"all_coms")

    #parameters for the various jobs
    self.sect_pars=self._option("sect_pars",err=1)

    #create the status object
    SEP.status.sep_status.__init__(self,self.logdir+"/job.stat",
      self.sect_pars.keys(), njob_extra=4,order=self._option("order",self.sect_pars.keys()))


    #shell to run to access the nodes
    self.shell=self._option("com_shell","rsh -n ")


    #add to command to restart the job;
    self.restart_com=self._option("restart_com","")


    #list of programs to be distributed
    self.prog_list=SEP.par_prog.distrib_list()
    

    #program to run ( simple parallelization)
    self.program=self._option("program")
    
    

    #number of attempts to start a job
    self.attempts=self._option("attempts",3)

    #global parameters (all jobs need these pars)
    self.global_pars=self._option("global_pars",SEP.par.sep_pars(name="par_job"))
    #machine object (later expand on the concept)
    self.mach =SEP.machine.sep_machine(self.logdir+"/job.mach",**self.kw)
 
    #maximum number of jobs
    self.njobs=int(self._option("njobs",2000))

    self.msg_objs=[] #object to receive messages from
    self.socket_objs=[] #socket objects

    #translation from machines to objects
    self.mlabel_talk={}

    #maximum number of processes that should use on socket
    self.nmax_talk=int(self._option("nmax_talk",60))

    #create the message object for rsh jobs
    self.msg_objs.append(msg_obj("%s/msgs.%d"%(self.logdir,0))) 
    self.first_port=int(self._option("port",50000))
    self.next_port=self.first_port
    for i in range(self.njobs/self.nmax_talk+1):
      self.msg_objs.append(server_msg_obj("%s/msgs.%d"%(self.logdir,i+1),
         self.nmax_talk,**self.kw))


  def _check_files(self,par_list):
    """ Check to make sure parallel files make sense

       Given a list of parallel files make sure that all of the files
       have the same number of sections (if not of type COPY)

    """
    nsect=len(par_list)
    self.mpi=SEP.mpi.mpi(logdir=self.logdir,verb=self.verb)
    self.multi_start=1
    for file in self.files.values():
      file.set_mpi_mach(self.mpi,self.mach)
      if file.type != "COPY":
        if nsect!=file.nsect():
          err = "Number of sections  %d not equal to number"
          err+= " of parts %d in file=%s" % (nsect,file.nsect,file.name)
          SEP.util.err(err)
      elif file.usage[0:3]=="OUT": self.multi_start=None

  def _collect(self):
     """Collect the data for all nodes where a process isn't currently running"""
     key_list=[]
     self._lock_all()
     mach_dict=self.get_other_dict(self.get_by_status("finished"))
     running=self.mach.get_by_status("running")
     running.extend(self.mach.get_by_status("sent"))
     running=self.mach.machs_from_labels(running)
     for key,val in mach_dict.items():
       if running.count(val) ==0 : key_list.append(key)
     if len(key_list)>0:
       time.sleep(5.)
       for file in self.files.values(): 
         if file.collect(key_list,[" "]):
           redo=str(self._option("redo_collect","n"))
           if redo[0]=="n" or redo[0]=="N" or redo[0]=='0': redo=None
           if redo:
             SEP.util.warn("Trouble collecting %s , redoing "%file.name)
             self._unlock_all()
             self.check_nodes()
             self._job_server()
             self._lock_all()
             if file.collect(key_list,[" "]):
               SEP.util.err("Trouble collecting %s  "%file.name)
           else: SEP.util.err("Trouble collecting %s  "%file.name)
     self._unlock_all()

  def _lock_all(self,read=1 ):
    """Lock all status files
       
       Make sure this is the only process that can update the status

    """
    self._lock_it(read)
    self.mach._lock_it()
    for name,file in self.files.items(): file._lock_it(read)

  def _unlock_all(self,write=1 ):
    """Unlock all status files"""
    self._unlock_it(write)
    self.mach._unlock_it()
    for file in self.files.values(): file._unlock_it(write)


  def update_parfile_dict(self,tag,parfile):
    """Modify the parfile dictionary

       When doing inversion the name of the model and datafile changes
       This routine allows for the change

    """
    self.files[tag]=parfile

  def set_global_par_dict(self,par):
    """Set the global database paramter list """
    self.global_pars=SEP.par.sep_pars(name="name")
    self.global_pars.add_pars(par)

  def set_global_par(self,par,value):
    """Set a parameter value to be used on all jobs"""
    self.global_pars.param(par,value)

  def _init_pids(self):
    """Mark that we are begining a new job"""
    self.key_pids={}
    self.pids=[]


  def _new(self):
    """Signify that this a new parallel job"""
    self._init_pids()
    self._lock_all(read=None)
    self.init_all()
    for msg in self.msg_objs: msg.init()
    for file in self.files.values(): file.init_all()
    self._unlock_all()
      

  def clean_files(self):
    """Clean the distributed executables and the distributed file parts"""
    self.mpi.clean()
    self.prog_list.clean()
    for file in self.files.values(): 
      if self.verb >1:  SEP.util.msg("Cleaning up %s"%file.name)
#      file.clean()
    
  def clean(self):
     """Clean up all files except output file(s)"""
     SEP.par_log.jobs.clean(self)
     for msg in self.msg_objs:  msg.clean();

  def restart(self):
    """Restart a job"""
    self._init_pids()
    self._logfile_update()
    self._msg_update()
    restart=SEP.status.sep_status.restart(self,1)
    for file in self.files.values(): file.restart(restart=1)

  def _logfile_update(self):
    """Update the message files based on  """
    self._lock_all()
    todo_list=self.get_elem_val(0,"running")
    some_progress=self.elem_not_list(["none"],4,keys=todo_list)
    log_files=self.get_stat_elem(4,some_progress)
    machs=self.get_stat_elem(1,some_progress)
    myre=re.compile("finished=(\d+)")
    line_list,err=self.msg_objs[0].read_delete()
    for key,file in log_files.items():
      mach=machs[key]
      f=open(file)
      lines=f.readlines()
      for line in lines:
        a=myre.search(line)
        if a:
          msg=encode_action("progress",key,mach,"finished=%s of"%str(a.group(1)))
          line_list.append("%s\n"%msg)
      f.close()
    self.msg_objs[0].write_msgs(line_list)
    self._unlock_all()
       

  def check_nodes(self,type="free"):
    """ Check the status of a series of nodes given by their machine label"""
    if self.verb >1:  SEP.util.msg("Checking node health")
    check=[]
    self._lock_all()
    machlabel=self.mach.get_by_status(type)
    for label in machlabel: 
      mach=self.mach.mach_from_label(label)
      if check.count(mach)==0:
        self.check_node_status(label)
        check.append(mach)
    self._unlock_all()


  def _init_run(self,restart):
    """Initialize a run. If restarting set restart=1 """
    if not restart: self._new()
    self.mach.init_all()
    self.check_nodes()
    self.failures=[]
    self.key_pids={}
    self.not_first=None
    self.key_finish={}
    for msg in self.msg_objs[1:]: msg.started=None
    for file in self.files.values():file.verb=self.verb
    SEP.spawn.run_wait("touch %s; rm %s"%(self.all_coms,self.all_coms),30)
    if restart: self.restart()
    self._check_files(self.sect_pars)

  def  run_job(self,restart=None,pars=None):
    """ Start a parallel job (restart=1 if restarting) """
    self._init_run(restart)
    stat=self.start_jobs()
    if stat== None: SEP.util.err("No jobs to run")
    self._lock_all()
    self.update_nrunning() 
    self._unlock_all()
    if stat=="Full":
      SEP.util.msg("All nodes are full. Will retry in 5 seconds")
      while "Full"==stat: time.sleep(5.)
      SEP.util.msg("Starting jobs")
    elif stat == "Finished" and self.nrunning==0:  self._collect()
    else:
      self._job_server()
      self._collect()

  def _msg_update(self):
   """Update the status using the message files"""
   self._lock_all()
   for i in range(len(self.msg_objs)-1,-1,-1):
     msg=self.msg_objs[i]
     if msg.started:
       lines,err=msg.read_delete()
       if err: 
         self.exit_clean(-1)
       for line in lines:
         oper,jobid,machid,extra=decode_action(line.rstrip("\n"))
         if   oper== 'record':  self._record(jobid,machid,extra)
         elif oper== 'finish':  self._finish(jobid,machid,extra)
         elif oper== 'progress':  self._progress(jobid,machid,extra)
         elif oper== 'error':  self._error(jobid,machid,extra)
         elif oper== 'thread':   self._slave_done(jobid,machid,extra)
         else: SEP.util.err("unknown type:%s:\n%s%s"%(oper,line,lines))
   self._unlock_all()

  def _del_socket_use(self,machid):
    """Mark that a given machine label is no longer using a socket"""
    #when restarting we want to do nothing so the if conditional
    if self.mlabel_talk.has_key(machid):
      self.msg_objs[self.mlabel_talk[machid]].remove_use(machid)
      del self.mlabel_talk[machid]
   

  def _add_socket_use(self,machid):
    """Find/create a socket for a given machine id to use"""
    for i in range(1,len(self.msg_objs)):
      good,pid,next_port=self.msg_objs[i].add_use(machid,self.next_port)
      if good:
        self.mlabel_talk[machid]=i
        if pid:
          self.pids.append(pid)
          self.next_port=next_port
        return i
    SEP.util.error("Ran out of msg_objects increase njobs and nmax_talk")

  def _slave_done(self,jobid,machid,extra):
    """Mark that a slave process has finished"""
    self._del_socket_use(machid)
    self.mach.update_status(machid,["free"])
    self.key_finish[jobid]=time.time()

  def _check_slave_status(self):
    """Check to see if a slave job  exited  incorrectly"""
    for key,myt in self.key_finish.items():
       if time.time() > myt+300: 
         stat=self.get_key_status_elem(key,0)
         mach=self.get_key_status_elem(key,1)
         extra=self.get_key_status_elem(key,2)
         if stat != "finished" and stat!="collected" and mach !="none":
           if extra != "seperr": extra="died"
           self._add_failure(key,mach,extra)
           os.waitpid(self.key_pids[key],0)
           del self.key_pids[key]
         del self.key_finish[key]

  def _check_status(self):
    """Check whether or not we have reached the criteria for quitting/finishing"""
    self._lock_all()
    self._check_slave_status()
    if self.get_by_status("fail"): self.exit_clean(-1)
    self.update_nrunning()
    self._unlock_all()

  def _record(self,jobid,machid,extra):
     """" Record that a task is running"""
     old_stat=self.get_key_status_elem(jobid,0)
     if old_stat=="sent":
       if self.verb >2: SEP.util.msg("  Job %s on %s registerd"%(jobid,machid))
       self.mach.update_status(machid,["running",jobid,extra]) 
       self.update_status(jobid,["running",machid])
       for file in self.files.values(): file.record(jobid)


  def _progress(self,jobid,machid,extra):
    """Register progress in a job"""
    old_stat=self.get_key_status_elem(jobid,0)
    old_extra=self.get_key_status_elem(jobid,2)
    src=re.compile("finished=(\d+)\s+")
    old=src.search(old_extra)
    new=src.search(extra)
    if old_stat!="finished" and old_stat!="collected":
      if not new or not old or int(old.group(1))< int(new.group(1)):
        if self.verb >1:
          SEP.util.msg("  Job %s on %s progress=%s"%(jobid,machid,extra))
        self.update_status(jobid,["running",machid,extra])

  def _finish(self,jobid,machid,extra):
     """ Mark that a job as finished"""
     for file in self.files.values(): file.done(jobid)
     if self.verb >1:
        SEP.util.msg("Job %s on %s finished"%(jobid,machid))
     self.update_status(jobid,["finished",machid])
     if self.key_finish.has_key(jobid): del self.key_finish[jobid]
#     os.waitpid(self.key_pids[jobid],0)
#     del self.key_pids[jobid]
     
  def prep_job(self,key,mlabel):
    """Do any prep work needed for job key to be run on mach label mlabel"""
    return None
    prog,mach=self.local_prog_mach(key,mlabel)
    if prog:
      return SEP.spawn.run_wait("rcp %s %s:%s"%(self.program,mach,
       prog),200,ntrys=5)
    return None
     
  def clean_job(self,key,mlabel):
    """Clean up after job (key) finished successfully"""
    return None
    prog,mach=self.local_prog_mach(key,mlabel)
     
  def _add_failure(self,key,mach,extra):
     """Error when running job"""
     self.set_failure(key,self.get_failures(key)+1)
     count=self.get_failures(key)
     self.failures.append(self._fail_label(key,mach))
     if count==self.attempts: 
       SEP.util.msg("To many failures for %s, dying"%key)
       self.exit_clean(-1)
     else: self.update_status(key,["todo"])
     #Future mark machine as failing????
     SEP.util.msg("Failure: key=%s mach=%s error=%s"%(key,mach,extra))
     self.check_node_status(mach)


  def check_node_status(self,mlabel=None,mach=None):
     """Check to see if a node is dead"""
     if mlabel: 
       mach=self.mach.mach_from_label(mlabel)
     elif not mach: SEP.util.err("Must provide mach or mlabel")
     stat=SEP.machine.check_alive(mach)
     if stat:
       SEP.util.msg("%s isn't functional (dead, mounting problem???)"%mach)
       labels=self.mach.label_list(mach)
       keys_on_node=[]
       key1=[]
       for m in labels:
         self.mach.set_key_status_elem(m,0,"dead")
         key1.extend(self.get_elem_val(1,m))
       #only deal with keys that aren't running (others will happen eventually)
       for  key in key1:
         stat=self.get_key_status_elem(key,0)
         if stat!="running" and stat!="sent": keys_on_node.append(key)
       self.handle_dead_jobs(keys_on_node)


  def handle_dead_jobs(self,labels):
    """Handle dead jobs on  a series of machine labels"""
    redo=str(self._option("redo","y"))
    if redo[0]=="n" or redo[0]=="N" or redo[0]=='0': redo=None
    for key in  labels:
      stat=self.get_key_status_elem(key,0)
      if stat!="collected":
        if stat!="finished" or redo: 
          self.set_key_status_elem(key,0,"todo")
          self.update_status(key,["todo","none","none","none"])
          for file in self.files.values(): file.redo(key)

     
  def update_nrunning(self):
    """Calculate how many jobs are currently running"""
    self.nrunning=len(self.sect_pars)-len(self.get_by_status("finished"))- len(self.get_by_status("collected"))

  def _error(self,id,machid,extra):
     """Error when running job"""
     self.update_status(id,["running",machid,"seperr"])
     return 1

  def _fail_label(self,key,mlabel):
    """return label for failures"""
    return "%s:%s"%(key,self.mach.mach_from_label(mlabel))

  def command_func(self,key,mlabel,par_args,job_args,tags):
    """Return the command to run given the job descriptor, job_string and file tags

       key      - the jobid 
       mlabel   - the machine label we will be running on
       par_args - the parameters for the parallel jobs
       job_args - the parameters for job args
       tags     - the local file tags

       This should be overwritten when doing more complex jobs

    """

    par_args.add_pars(job_args)
    if not self.program: 
      SEP.util.err("Must override command_func when program not provided")
    prog=self.prog_list.prepare(self.program,self.mach.mach_from_label(mlabel))
    run ="%s sep.begin_par=1 sep.end_par=1 %s %s"%(prog,
      string.join(par_args.return_sep_par_list()),
     string.join(tags))
    return run

  def return_par_files(self):
    """Return the parallel files"""
    return self.files 

  def __fix_rsh_com(self,mach,inp):
    """Add the shell command to a job 

         mach - machine we are running
         inp  - input command string
        
         return the fixed command line

    """
    list=inp.split()
    for c in list: 
      if c == '"': c="'"
    out=self.shell+" "+mach+' "'+string.join(list)+'"'
    return out

  def set_failure(self,key,val):
     """Set the number of failures for a given job"""
     self.set_key_status_elem(key,3,val)
                                                                                
                                                                                
  def status_vals(self,keys):
     """Return the status of a series of keys"""
     return self.get_stat_elem(0,keys)

  def get_other_dict(self,key_list):
     """Get the other status given a key list"""
     return self.get_stat_elem(1,key_list)
                                                                                
  def get_failures(self,key):
     """Get the the number of failures of a given job"""
     v= self.get_key_status_elem(key,3)
     if not v or v == "none": v=0
     else: v=int(v)
     return v

  def start_jobs(self):
    """ Start a series of jobs """
    commands,err=self.build_commands()
    start_dt=1.5
    if self.not_first: isleep=-start_dt
    else:              
      isleep=2-start_dt
      self.not_first=1 
    if err == "Run":
      for key,cm in commands.items(): 
         log=self.logfile(label=key,old=1)[0]
         SEP.spawn.run_wait("echo %s >>%s"%(cm,self.all_coms),40)
         mlabel=self.get_key_status_elem(key,1)
         pid=os.fork()
         isleep=isleep+start_dt
         if mlabel == "none": SEP.util.err("Trouble getting key label %s"%key)
         if not pid: #child process starts jobs
           time.sleep(isleep)
           a= self.prep_job(key,mlabel)
           if a:
             self.msg_objs[0].action("error:%s:%s:prep failure"%(key,mlabel))
           self.wait(cm,verb=4,error=None,label=key,logfile=log)
           self.msg_objs[0].action("thread:%s:%s:none"%(key,mlabel))
           self.clean_job(key,mlabel)
           sys.exit(0)
         else: 
           self.key_pids[key]=pid
    return err



  def _job_server(self):
     """Update node status and try to start more jobs"""
     sleep_time=self.status_interval
     next_time=time.time()+self.alive_interval
     sent_list=[]
     while 1:
       time.sleep(sleep_time)
       self._check_status()
       self._msg_update()
       stat=self.start_jobs()
       if self.nrunning==0: 
         self._lock_all()
         self.exit_clean()
         self._unlock_all()
         return 
       if time.time()>next_time:
         self.check_nodes("running")
         self._lock_all()
         next_time=time.time()+self.alive_interval
         dlist=self.mach.get_by_status("dead")
         if dlist:
           mlist=[]
           for mach in dlist: mlist.append(self.mach.mach_from_label(mach))
           mlist=SEP.util.condense(mlist)
           for mach in mlist:
             stat=SEP.machine.check_alive(mach)
             if not stat:
               SEP.util.msg("%s now seems functional"%mach)
               labels=self.mach.label_list(mach)
               for m in labels: self.mach.set_key_status_elem(m,0,"free")
         self._unlock_all()
         
  def assign_jobs(self):
     """Return the list of jobs to assign to each node """
     todo_list=self.get_by_status("todo")
     if self.multi_start:
       machlabel=self.mach.free_labels()  #get list of machines we can run on
     else:
       machlabel=self.mach.free_mach()  #get list of machines we can run on
     mach_send={}
     for mlabel in machlabel:
       assign=0
       i=0
       while assign ==0 and  i < len(todo_list):
          flabel=self._fail_label(todo_list[i],mlabel)
          if self.failures.count(flabel)==0:
            assign=1 
            mach_send[todo_list[i]]=mlabel
            del todo_list[i]
          else: i=i+1
     return mach_send



  def server_pars(self,mach,job_pars):
    """Return the socket parameters need for a job"""
    icom=self._add_socket_use(mach)
    return self.msg_objs[icom].server_pars(job_pars)

           

  def build_commands(self):
    """Build commands to run on free nodes"""

    self._lock_all()
    mach_send=self.assign_jobs()
    if not mach_send: 
      if self.get_by_status("todo"): 
        self._unlock_all()
        return None,None
      if self.get_by_status("running"): 
        self._unlock_all()
        return None,None
      self._unlock_all()
      return None,"Finished"

     
    #setup all of the parralel files for the job
    tag_job={}
    for key in mach_send.keys(): tag_job[key]=" ";

    for file in self.files.values():
#      if self.verb >1: 
#        SEP.util.msg("  Distributing %s to  %s"%(file.name,
#           string.join(mach_send.values())))
      err,tags=file.tags(mach_send)  #tupple  label:tag
      if err: 
        for mach in mach_send.values(): self.check_node_status(mach)
        self._unlock_all()
        return None,None
      for key,tag in  tags.items(): tag_job[key]=tag_job[key]+" "+tag

    #time to build the commands
    run_this={}
    for key,mach in mach_send.items():
      self.update_status(key,["sent",mach_send[key]])#update our status
      self.mach.update_status(mach,["sent",key,str(time.time())])  #update machine status
      job_pars=SEP.par.sep_pars(seppar=self.global_pars,name=key)
      job_pars=self.server_pars(mach,job_pars)
      job_pars.add_param("sep.jobid",key)
      job_pars.add_param("sep.mach_label",mach_send[key])
      job_args=SEP.par.sep_pars(seppar=self.sect_pars[key],name="sect_par")
      self.set_key_status_elem(key,4,self.logfile(label=key)[0])
      stat=self.get_key_status_elem(key,2)
      srch=re.compile("(.+)=(.+)")
      if self.restart_com: g=srch.search(self.restart_com)
      else:g=None
      if g: 
        if job_pars.param(g.group(1)):
          job_pars.del_par(g.group(1))
      if stat and stat != "none":
        job_args.add_string(stat)
        if self.restart_com:
          job_args.add_string(self.restart_com)
      if self.verb >0:
        if self.verb ==1 or self.verb==2:
          SEP.util.msg("Sending job %s to %s "%(key,mach))
        elif self.verb ==3:
          SEP.util.msg("Sending job %s to %s pars=[%s]"%(key,mach,self.sect_pars[key].return_sep_par_string()))
        else:
          SEP.util.msg("Sending job %s to %s pars=[%s]"%(key,mach,string.join(job_pars.return_sep_par_list())))
      run=self.command_func(key,mach,job_pars,job_args,tag_job[key].split())
      run_this[key]=self.__fix_rsh_com(self.mach.mach_from_label(mach), run)
    self._unlock_all()
    return run_this,"Run"

  def exit_clean(self,stat=None):
    """Kill (or exit) all slave threads"""

    #first terminate all of the jobs
    for key,pid in self.key_pids.items():
      if stat: os.kill(pid,9)
#      else: os.waitpid(pid,0)


    while 1:
       pid, sts = os.waitpid(-1, os.WNOHANG)
       if pid == 0: break
    #exit all the socket servers
    self.exit_servers()


#    #now make sure that all of the sockets are dead
#    for pid in self.pids: 
#      os.waitpid(pid,0)

    if stat: sys.exit(stat)

  def exit_servers(self):
    """Exit all of the socket servers"""

    ip=self.msg_objs[1].ip_server()
    port_nums=[]
    for i in range(1,len(self.msg_objs)):
      if self.msg_objs[i].started: 
        port_nums.append(str(self.msg_objs[i].port_number()))

    while 1:
      mach=self.mach.not_dead()
      prog="%sSend_msg.%s.%s"%(SEP.datapath.datapath(mach),
         pwd.getpwuid(os.getuid())[0],os.getpid())
      err=SEP.spawn.run_wait("rcp %s/Send_msg %s:%s"%(SEP.paths.sepbindir,
        mach,prog),200 ,quiet=1,ntrys=5)
      if err: self.check_node_status(mach=mach)
      
      else:
        err1=SEP.spawn.run_wait("%s %s '%s ip=\"%s\" port=%s msg_type=%s %s'"%(self.shell,
        mach, prog,ip,string.join(port_nums,","),"exit",
         "</dev/null >/dev/null"),30, error=0)
        err2=SEP.spawn.run_wait("%s %s 'rm -f %s'"%(self.shell,mach,prog),
          30,error=0,quiet=1)
        if not err1 and not err2: return

  def _option(self,par,default=None,err=None):
    """Simple parameter checking

       par     - parameter to check for 
       default - default parameter value
       err     - whether or not to give an error if the parameter doesn't exist

    """
    if self.kw.has_key(par): return self.kw[par]
    elif default: return default
    elif err: SEP.util.err("Missing required parameter %s "%par)
      


class par_job(parjob,SEP.options.par_group,SEP.prog.job):
  def __init__(self,name="PAR"):
    """Initialize a parallel object """
    SEP.options.par_group.__init__(self,name)
    SEP.prog.job.__init__(self,name)
    self.add_doc_params()

  def add_doc_params(self):
     """Add the required files for a paralllel job"""
     self.add_doc_parameters(_cl_par_params_list())
     self.add_doc_parameters(_object_par_params_list())


  def runtime_check_build(self):
     """Check to make sure the parameters are reasonable/add additional parameters"""
     self.kw=self.return_args().return_dict()
     self.kw["files"]=self.param("files")
     self.kw["sect_pars"]=self.param("sect_pars")
     self.kw["order"]=self.param("order")
     self._init_check()

  def delete_cl_par_params(self):
    """Delete the command line parameters of the parallel job"""
    for i in _cl_par_params_dict().keys(): self.del_doc_param(i)

  def delete_object_par_params(self):
    """Delete the object parameters of the parallel job"""
    for i in _object_par_params_dict().keys(): self.del_doc_param(i)


  def set_cl_par_params(self,pars):
    """Set the commandline par parameters from an SEP.par"""
    for i in _cl_par_params_dict().keys(): 
      self.add_param(i,pars.param(i))
    
     

class par_job_nfs(par_job):
  """ A parallel job object where we are sharing image files"""

  def __init__(self,name="PAR"):
    """Initialize a parallel job object where we sharing image files

       SEE SEP.par_job.parjob

    """
    par_job.__init__(self,name)

  def  runtime_check_build(self):
    """Add additional parameters and check to see if the parameters make sense"""
    par_job.runtime_check_build(self)
    self.registered=None
    self.kw['redo']=None
     
  def add_doc_params(self):
    """Add the required files for a nfs paralllel job"""
    parjob.add_doc_param()
    self.del_doc_param("program")
    self.del_doc_param("redo")

  def assign_jobs(self):
     """Return the list of jobs to assign to each node 
        In this case we will only allow one to be started at a time
        to make sure NFS read/write doesn't kill us"""
     mach_send=parjob.assign_jobs(self)
     mach_={}
     if mach_send and not self.registered: 
       self.registered=1
       mach_[mach_send.keys()[0]]=mach_send.values()[0]
     return mach_
       

  def _record(self,jobid,machid,extra):
     """" Record that a task is running"""
     parjob._record(self,jobid,machid,extra)
     self.registered=None




def _object_par_params_list():
   """Return the object parallel parameters that are usually specified by the inheritance class"""
   list=[]
   list.append(SEP.options.parameter("nmax_talk",60,"Maximum number of processes to use a socket"))
   list.append(SEP.options.parameter("files","required","Dictionary of a parallel files"))
   list.append(SEP.options.parameter("sect_pars","required","Dictionary jobid:parameters"))
   list.append(SEP.options.parameter("program","none","Parallel program to run (required if not overriding)"))
   list.append(SEP.options.parameter("global_pars","none","parameters that every job must have"))
   list.append(SEP.options.parameter("order","none","Order to do jobs (semi-random)"))
   return list


def _cl_par_params_list():
   """Return the basic parallel parameters that are usually specified on the commandline"""
   list=[]
   list.append(SEP.options.parameter("nmax_talk",60,"Maximum number of processes to use a socket"))
   list.append(SEP.options.parameter("device","eth1","Device connected to the cluster"))
   list.append(SEP.options.parameter("port",50000,"First port to look for an open connection"))
   list.append(SEP.options.parameter("redo","y","Whether or not to redo a job on a dead node"))
   list.append(SEP.options.parameter("redo_collect","n","Whether or not to redo a job if a node dies during collect"))
   list.append(SEP.options.parameter("status_interval",10.,"sleep time in job loop (between checking message files)"))
   list.append(SEP.options.parameter("alive_interval",600.,"how often to check if the nodes are working correctling"))
   list.append(SEP.options.parameter("pverb",0,"Verbosity level for parallel operatorions",required=None))
   list.append(SEP.options.parameter("logdir","stat","the directory we send logging messages into"))
   list.append(SEP.options.parameter("shell","rsh -n","the shell command to use when starting remote jobs"))
   list.append(SEP.options.parameter("njobs",2000,"The maximum number of jobs to un at one time"))
   list.append(SEP.options.parameter("attempts",3,"The number of attempts to run a job"))
   return list 

def _cl_par_params_dict():
   """Return the basic parallel parameters (command line) as a dictionary"""
   dict={}
   for i in _cl_par_params_list(): dict[i.param]=i
   return dict

def _object_par_params_dict():
   """Return the basic parallel parameters (objects) as a dictionary"""
   dict={}
   for i in _object_par_params_list(): dict[i.param]=i
   return dict

class parameters(SEP.options.par_group):
  """A class the holds the basic parallel job parameters"""
  def __init__(self,name="PARJOB"):
    SEP.options.par_group.__init__(self,name)
    self.add_doc_params()
  def add_doc_params(self):
    self.add_doc_parameters(SEP.parjob._cl_par_params_list())
  def run_job(self,restart,a=None):
    """Nothing to run for this class"""



