import SEP.vector
import SEP.util
import SEP.operator
import SEP.status
import re,sys

__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2005.2.15"
                                                                                



class solver:
  """A generic solver object. 

     All other linear solvers are derived from this class

   """

  def __init__(self,name,x,resid,op,stepper,niter,verb=0,restart=None):
    """Initialze a generic solver object

        name    - the name for the solver
        x       - the model vector
        resid   - the residual vector
        op      - the operator we are inverting
        stepper - the stepper object
        niter   - the number of iterations
        verb    - whether or not to be verbose
        restart - whether or not we are restarting 

    """
    self.x=x
    self.rr=resid
    self.op=op
    self.niter=niter
    self.verb=verb
    self.iter0=0
    self.name=name
    self.stepper=stepper
    self.g=self.x.clone_space(name=self.name+"_g")
    self.gg=self.rr.clone_space(name=self.name+"_gg")
    if not restart:
      self.g.zero()
      self.gg.zero()
    else:
      self.g.load()
      self.gg.load()
    if self.verb:  self.dd=self.rr.dot(self.rr).real/1000. 
    self.vect_list.extend([self.g,self.gg])
    self.stepper.alloc(x,resid,restart)
    job_list=self.calc_work()
    self.status=SEP.status.sep_status(self.name+".stat",job_list)
    self.restart=restart
    self.op.init_op(restart)
    self.restart_it()

  def clean(self):
    """Clean the solver """
    self.stepper.clean()
    for vec in self.vect_list: 
      vec.clean()
    self.op.clean()

  def restart_it(self):
    """Restart the solver"""
    #update status
    restart_list=self.status.restart(self.restart,["finished"],["started","restart"])
    if restart_list:
      iterS=re.compile("^(\d+)\.").search(restart_list[0])
      if  not iterS: SEP.util.err("Unknown status parameter %s"%restart_list[0])
      iter=int(iterS.group(1))
      self.iter0=iter+1
      if re.compile("_a$").search(restart_list[0]): step=0
      elif re.compile("_f$").search(restart_list[0]): step=1
      elif re.compile("step$").search(restart_list[0]): step=2
      elif re.compile("extra$").search(restart_list[0]): step=3
      else: SEP.util.err("unknown parameter %s"%restart_list[0]);
      if iter >0: self.stepper.set_forget(None)
      if step==0: 
        self.op.adjoint(self.g,self.rr,self.status,iter,restart=restart_list[0])
        self.op.forward(self.g,self.gg,self.status,iter)
      elif step==1: 
        self.op.forward(self.g,self.gg,self.status,iter,restart=restart_list[0])
      if step <3: 
        if not self.step(iter): return None
      if not self.extra(iter): return None

  def calc_work(self):
    """Calculate what has to be done for the inversion 

       This information is used for restarting

        basically makes a list of the form

        op.0.adj
        op.0.forward
        op.0.cgstep
        ....

    """
    job_list=[]
    for iter in range(self.niter):
      f,a=self.op.job_desc(iter)
      job_list.extend(a)
      job_list.extend(f)
      job_list.append(str(iter)+".step")
      job_list.append(str(iter)+".extra")
    return job_list


  def step(self,iter):
    """Do a step"""
    self.status.update_status(str(iter)+".step",["started"],lock=1)
    if not self.stepper.step(self.x,self.g,self.rr,self.gg): return None
    self.status.update_status(str(iter)+".step",["finished"],lock=1)
    return 1

  def extra(self,iter):
    """Allow additional calculations at the end of an iteration"""
    self.status.update_status(str(iter)+".extra",["started"],lock=1)
    self.end_iter(iter)
    self.status.update_status(str(iter)+".extra",["finished"],lock=1)
    return 1

  def end_iter(self,iter):
    """Do any additional work at the end of an iteration"""
  def report(self,iter):
    """Report the energy in varios vectors"""
    dd=self.rr.dot(self.rr)
    dx=self.x.dot(self.x)
    dg=self.g.dot(self.g)
    if iter==0: 
      self.dx=dx.real/1000. 
      self.dg=dg.real/1000. 
    SEP.util.msg( "iter=%d r(dat)=%f r(x)=%f r(grad)=%f"%(iter,dd.real/self.dd,
      dx.real/self.dx,dg.real/self.dg))

  def iter(self):
    """Perform one iteration of the iterative solver"""
    for iter in range(self.iter0,self.niter):
      self.op.adjoint(self.g,self.rr,self.status,iter)
      self.op.forward(self.g,self.gg,self.status,iter)
      if not self.step(iter): return None
      if not self.extra(iter): return None
      if self.verb>0: self.report(iter) 
    return 1

class smp(solver):
  """A  class for simple inversions d= Lm"""
  def __init__(self,name,m,d,op,stepper,niter,wop=None,
       m0=None,resd=None,verb=None,restart=None):
    """Initialize a simple inversion operator
     
      name   - is the name associated with the solver object
      m      - is the model vector
      d      - is the data vector
      op     - is the operator
      stepper- is the step function object
      niter  - is the number of iterations
      wop    - is the weight operator (None)
      m0     - is the initial value for the model (0)
      resd   - is the initial value for the data residual (0)
      verb   - whether or not to be verbose
      restart- whether or not to restart
       

    """
    if wop:
      self.rr=wop.range.clone_space(name=name+".smp_rr")
      if restart: self.rr.load()
      else:
        self.rr.zero()
        wop.forward(d,self.rr)  #rr= W d
        self.rr.scale(-1.)      #rr= - Wd
      self.chain_op=SEP.operator.chain(name+".WL",[op,wop],restart,verb=verb)
    else:
      self.rr=d.clone_space(name=name+".smp_rr") # rr=d
      if  restart: self.rr.load()
      else:
        self.rr.zero()
        self.rr.scale_add(-1.,d)   # rr=-d 
        self.chain_op=op
    self.vect_list=[self.rr]

    if m0 and not restart:
      m.zero()
      m.add(m0)
      self.chain_op(m,self.rr,add="yes")

    self.m=m
    solver.__init__(self,name,self.m,self.rr,self.chain_op,stepper,niter,verb,restart)

      

  
  def solve(self):
    """Solve for the model"""
    if not self.iter(): return None
    return 1

class reg(solver):
  """Perform regularaized inversion"""

  def __init__(self,name,m,d,op,reg,stepper,niter,eps,wop=None,rm0=None,
       m0=None,resd=None,verb=None,restart=None):
    """Initialize a simple inversion operator
     
      name   - is the name associated with the solver object
      m      - is the model vector
      d      - is the data vector
      op     - is the operator
      reg    - is the regularization operator
      stepper- is the step function object
      niter  - is the number of iterations
      eps    - is the value for epsion (relative weighting)
      wop    - is the weight operator (None)
      m0     - is the initial value for the model (0)
      resd   - is the initial value for the data residual (0)
      verb   - whether or not to be verbose
      restart- whether or not to restart
       

    """

    self.status=SEP.status.sep_status(name+".stat",name)

    if wop:
      self.rd=wop.range.clone_space(name=name+"_rd")
      if restart: self.rd.load()
      else: 
        self.rd.zero()
        wop.forward(d,self.rd)       #rd= W d
        self.rd.scale(-1.)      #rd= - Wd
      self.chain_op=SEP.operator.chain(name+".WO",[op,wop],restart=restart,verb=verb)
    else:
      self.rd=d.clone_space(name=name+"_rd") # rd=d
      if restart: self.rd.load()
      else: 
        self.rd.zero()  
        self.rd.scale_add(-1.,d) #rd=-d
      self.chain_op=op

    self.rm=reg.range.clone_space(name=name+"_rm")
    if  restart: self.rm.load()
    else:
      self.rm.zero()
      if rm0: self.rm.add(rm0)
    self.vect_list=[self.rm,self.rd]

    self.scale_op=SEP.operator.scale("eps_reg",m,m,eps,verb=verb)
    self.reg_chain=SEP.operator.chain("eA",[reg,self.scale_op],restart=restart,verb=verb)
    self.m=m

    if not restart: self.m.zero()
    if m0 and not restart:
      self.m.add(m0)    # m=m0
      self.chain_op(self.m,self.rd,add=1)
      self.reg_chain(self.m,self.rm,add=1)
      
    self.rr=SEP.vector.super(name+".rr",[self.rd,self.rm])
    self.op=SEP.operator.array(name+".main",[self.chain_op,self.reg_chain],1,2,restart=restart,verb=verb)
    solver.__init__(self,name,self.m,self.rr,self.op,stepper,niter,verb,
     restart)

  def solve(self):
    """Iterate the problem"""
    if not self.iter(): return None
    return 1

class prec(solver):
  """Perform model preconditioning regularized inversion"""
  def __init__(self,name,m,d,op,prec,stepper,niter,eps,wop=None,rm0=None,
       p0=None,resd=None,verb=None,restart=None):
    """Initialize a regularization with model preconditioning inversion operator
     
      name   - is the name associated with the solver object
      m      - is the model vector
      d      - is the data vector
      op     - is the operator
      prec   - is the preconditioning operator
      stepper- is the step function object
      niter  - is the number of iterations
      eps    - is the epsilon value (relative weighting)
      wop    - is the weight operator (None)
      p0     - is the initial value for the preconditioned variable (0)
      resd   - is the initial value for the data residual (0)
      verb   - whether or not to be verbose
      restart- whether or not to restart
       

    """


    if wop:
      self.rd=wop.range.clone_space(name=name+"_rd")
      if restart: self.rd.load()
      else: 
        self.rd.zero()
        wop.forward(d,self.rd)       #rd= W d
        self.rd.scale(-1.)      #rd= - Wd
      self.chain_op=SEP.operator.chain(name+".WLP",[prec,op,wop],restart=restart,verb=verb)
    else:
      self.rd=d.clone_space(name=name+"_rd")
      if restart: self.rd.load()
      else: 
        self.rd.zero()
        self.rd.scale_add(-1.,d)   # rd=-d 
      self.chain_op=SEP.operator.chain(name+".LP",[prec,op],restart=restart,verb=verb)

    self.rm=prec.domain.clone_space(name=name+"_rm")
    if restart: self.rm.load()
    else:
      self.rm.zero()
      if rm0: self.rm=rm0.add(rm0)

    self.prec=prec
    self.m=m
    if not restart: self.m.zero()

    self.p=prec.domain.clone_space(name=name+"_p")
    if restart: self.p.load()
    else:
      self.p.zero()
      if p0: self.p.add(p0)
      
    self.scale_op=SEP.operator.scale(name+".eps",self.p,self.m,eps,restart=restart,verb=verb)
      

    self.vect_list=[self.rd,self.rm,self.p]
    self.rr=SEP.vector.super(name+".rr",[self.rd,self.rm])
    self.op=SEP.operator.array(name+".main",[self.chain_op,self.scale_op],1,2,restart=restart,verb=verb)
    solver.__init__(self,name,self.p,self.rr,self.op,stepper,niter,verb,restart)

  def calc_work(self):
    """Calculate what has to be done for the inversion 

       This information is used for restarting

        basically makes a list of the form

        op.0.adj
        op.0.forward
        op.0.cgstep
        ....

    """
    job_list=solver.calc_work(self)
    f=self.prec.job_desc(-1)[0]
    job_list.extend(f)
    return job_list

  def solve(self):
    """Solve for a model"""
    if not self.iter(): return None
    self.prec.forward(self.p,self.m,self.status,-1)
    return 1
