import SEP.options
import SEP.par
import SEP.sepfile
import SEP.operator
import SEP.sep_vector
import SEP.parjob
import SEP.parfile
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2004.12.16"
                                                                                


"""A module for dealing with WEI programs"""


class par(SEP.options.par_group):
  """The base class for parameters necessary to run WEI programs"""

  def __init__(self,name="WEI"):
     """Initialize the base class for WEI programs"""
     SEP.options.par_group.__init__(self,name)
     self.add_doc_params()
     self.data_des=None
     self.model_des=None
     self.datum_des=None

  def add_doc_params(self):
    """Add  parameter info"""
    self.add_doc_param("verb",doc="Verbosity level",required=None)
    self.add_doc_param("first_dz",doc="First dz step",required=None)
    self.add_doc_param("velocity",'y',"Velocity (y) or slowness (n)")
    self.add_doc_param("npad_y",'0',"Padding in y to kill edge effects")
    self.add_doc_param("npad_x",'0',"Padding in x to kill edge effects")
    self.add_doc_param("program",doc="WEI program to run")
    self.add_doc_param("R",doc="Image (mx,my,hx,hy,w)")
    self.add_doc_param("S",doc="Slowness file (x,y,z)")
    self.add_doc_param("D",doc="Data file (mx,my,hx,hy,w)")
    self.add_doc_param("schoice",doc="Slowness reference (nfef,2,nz)")
    self.add_doc_param("nzb",
     doc="Size of depth block (defaults to 100MB worth)",required=None)
    self.add_doc_param("nwb",1,doc="Size of frequency block")

  def datumsep(self):
    """Return the sepfile description for the data"""
    if not self.datum_des:
      self.datum_des=SEP.sep_vector.scmplx(tag=self.param("DOWN"))
    return self.datum_des

  def dsep(self):
    """Return the sepfile description for the data"""
    if not self.data_des:
      self.data_des=SEP.sep_vector.scmplx(tag=self.param("D"))
    return self.data_des

  def rsep(self):
    """Return the sepfile description for the image"""
    if not self.model_des:
      d=SEP.sepfile.sep_file(self.param("R"))
      esize=d.history.param("esize",4)
      if esize==4:
        self.model_des=SEP.sep_vector.sfloat(tag=self.param("R"))
      else:
        self.model_des=SEP.sep_vector.scmplx(tag=self.param("R"))
    return self.model_des

  def set_dsep(self,d):
    """Set the sepfile description for the data"""
    self.data_des=d

  def set_rsep(self,r):
    """Set the sepfile description for the image"""
    self.model_des=r
   
  def set_datumsep(self,r):
    """Set the sepfile description for the datumed data"""
    self.datum_des=r
   

  def _check_offset(self):
     """Check to make sure we have 0 offset in x and y"""
     n=int(self.param("ahx_n"))
     o=float(self.param("ahx_o"))
     d=float(self.param("ahx_d"))
     f=-float(o)/float(d)
     i=int(f+.5)
     if abs(f-i) > abs(d)/100.:
       print i,f,d
       SEP.util.err("Offset x doesn't contain 0 offset in hx")
     n=int(self.param("ahy_n"))
     o=float(self.param("ahy_o"))
     d=float(self.param("ahy_d"))
     if n>1:
       f=-float(o)/float(d)
       i=floor(f+.5)
       if abs(f-i) > abs(d)/100.:
         SEP,util.err("Offset y doesn't contain 0 offset in hy")

  def _set_nzb_nwtot(self):
    """Set nzb to a reasonable value"""
    n1=int(self.param("amx_n") )
    n2=int(self.param("amy_n") )
    n3=int(self.param("ahx_n") )
    n4=int(self.param("ahy_n") )
    n5=int(self.param("az__n") )
    nw=int(self.param("aw__n") )
    sz = int(100*1024*1024/n1/n2/n3/n4)
    if sz < 1: sz==1
    if sz > n5: sz=n5
    nzb=self.param("nzb")
    if not nzb: self.add_param("nzb",sz)
    self.add_param("nw_tot",nw)
    

  def build_check_params(self):
    """Check to make sure parameters make sense"""
    if self.param("program") == "none":
      SEP.util.err("Must specify WEI program")


  def runtime_check_build(self):
    """Make sure everything looks right before running"""
    self._check_slowness()
    self._check_schoice()

  def create_down(self,name):
    """Create the data space with the file name name"""
    return self._create_file(name,["mx","my","hx","hy","w_"])

  def create_data(self,name):
    """Create the data space with the file name name"""
    return self._create_file(name,["mx","my","hx","hy","w_"])

  def _create_file(self,name,axes,do_real=None):
    """Create  a file using the axis description in axes

       name - the name of the file to create
       axes - the list of axes to use (["mx","my","hx","hy","w_"]
       do_real- wheter of not the data is real
                                                                                           
    """
    if do_real: junk=SEP.sep_vector.sfloat(name=name,space_only=1)
    else:       junk=SEP.sep_vector.scmplx(name=name,space_only=1)
    for i in range(len(axes)):
       n,o,d=self._get_wave_axis(axes[i])
       junk.set_axis(i+1,n,o,d,axes[i],"none")
    if do_real:junk.history.add_param("esize",4)
    else:junk.history.add_param("esize",8)
    return junk

  def _get_wave_axis(self,par):
    """Return aX_n, aX_o, aX_d"""
    n=self.param("a%s_n"%par)
    o=self.param("a%s_o"%par)
    d=self.param("a%s_d"%par)
    return n,o,d

  def _check_slowness(self):
    """Check to make sure slowness is logical"""
    S=SEP.sepfile.sep_file(self.param("S"))
#    if S.ndims()!=3: SEP.util.err("Expecting slowness to be 3-D")
    n,o,d,label,unit=S.axis(3)
    nz=int(self.param("az__n"))
    oz=float(self.param("az__o"))
    dz=float(self.param("az__d"))
    if n!=nz or abs((oz-o)/dz) >.001 or abs(1.-d/dz) >.001:
      SEP.util.err("Expecting sampling in slowness and az_[n,o,d] to be identical")
    d_z= self.param("first_dz",error=None)
    if not dz: self.add_param("first_dz",dz)
    self.check_slow_extend()


  def check_slow_extend(self):
    """Check to see if the slowness needs to be extended"""
    S=SEP.sepfile.sep_file(self.param("S"))
    self._slow_extend_check(S,1,"x")
    self._slow_extend_check(S,2,"y")
    self.extend_check=1
   
  def _slow_extend_check(self,ssep,iax,ax):
    """Check how much we need to extend slowness and report

       ssep  - Slowness file
       iax   - axis of slowness file
       ax    - axis name
    
    """
    nslow,oslow,dslow,label,unit=ssep.axis(iax)
    if nslow==1: return
    slow_min,slow_max=self._min_max(nslow,oslow,dslow)
    m_min,m_max=self._min_max(int(self.param("am%s_n"%ax)),float(self.param("am%s_o"%ax)),
             float(self.param("am%s_d"%ax)))
    h_min,h_max=self._min_max(int(self.param("ah%s_n"%ax)),float(self.param("ah%s_o"%ax)),
             float(self.param("ah%s_d"%ax)))
    if abs(h_min) >abs(h_max): ln=abs(h_min)
    else: ln=abs(h_max)
    if m_min - ln < slow_min:
      SEP.util.msg("Slowness will be extended by %f in %s below axis"%(slow_min-m_min+ln,
        str(ax)))
    if m_max + ln > slow_max:
      SEP.util.msg("Slowness will be extended by %f in %s above axis"%(m_max+ln-slow_max,
        str(ax)))

  def _min_max(self,n,o,d):
    """Return min and max based on n, o, and d"""
    b=o
    e=o+d*(n-1)
    if b < e: return b,e
    return e,b


  def _check_schoice(self):
    """Check to make sure schoice is logical"""
    nz=int(self.param("az__n"))
    schoice=SEP.sepfile.sep_file(self.param("schoice"))
#    if schoice.ndims()!=3:
#      SEP.util.err("Expecting schoice to be 3-D")
    n2,o,d,label,unit=schoice.axis(2)
    n3,o,d,label,unit=schoice.axis(3)
    if n2!=2 or n3!=nz:
      SEP.util.err("Expecting schoice n2=2 n3=nz")

  def add_mid(self,tag):
    """Add the midpoint information based on a tag"""
    ss=SEP.sepfile.sep_file(self.param(tag))
    self._axis_add(ss,"mx",1)
    self._axis_add(ss,"my",2)

  def add_off(self,tag):
    """Add the midpoint information based on a tag"""
    ss=SEP.sepfile.sep_file(self.param(tag))
    self._axis_add(ss,"hx",3)
    self._axis_add(ss,"hy",4)

  def add_ph(self,tag):
    """Add the midpoint information based on a tag"""
    ss=SEP.sepfile.sep_file(SEP.param(tag))
    ss._axis_add(ss,"px",3)
    ss._axis_add(ss,"py",4)
       
  def _axis_add(self,wspace,pa,iax):
     """Add an axis to parameter list
                                                                                           
        wspace - sepfile containing the wavespace dimensions
        pa     - the axis descriptor  in WEI
        iax    - the axis in the dataset
                                                                                           
     """
     n,o,d,label,unit=wspace.axis(iax)
     self.add_string("a%s_n=%d a%s_o=%f a%s_d=%f"%(pa,int(n),
      pa,float(o),pa,float(d)))
                                                                                           
#######################################################
#######################################################
#######################################################
class datum_par(par):
  """Class for doing offset based WEI"""
  def __init__(self,name="Datuming WEI"):
     """Iniatialize offset WEI parameter class"""
     par.__init__(self,name)
  def add_doc_params(self):
     """Add documentation specific to WEI  in the offset domain """
     par.add_doc_params(self)
     self.del_doc_param("R")
     self.add_doc_param("DOWN",doc="Datumed dataset")
     self.add_doc_param("az__n",doc="Number of depth steps")
     self.add_doc_param("az__o",doc="First depth")
     self.add_doc_param("az__d",doc="Sampling in depth")
     
  def build_check_params(self):
     """Build the parameters for offset based WEI"""
     par.build_check_params(self)
     self.add_mid("D")
     self.add_off("D")
     self._axis_add(SEP.sepfile.sep_file(self.param("D")),"w_",5)
     self.add_param("ray_par","n")
     self._check_offset()
     self._set_nzb_nwtot()
     self.set_datumsep(self.create_data("junk"))

#######################################################
#######################################################
#######################################################
class off_par(par):
  """Class for doing offset based WEI"""
  def __init__(self,name="Offset WEI"):
     """Iniatialize offset WEI parameter class"""
     par.__init__(self,name)
  def add_doc_params(self):
     """Add documentation specific to WEI  in the offset domain """
     par.add_doc_params(self)
  def build_check_params(self):
     """Build the parameters for offset based WEI"""
     par.build_check_params(self)
     self._axis_add(SEP.sepfile.sep_file(self.param("D")),"w_",5)
     self.add_param("ray_par","n")
     self._check_offset()
     self._set_nzb_nwtot()

  def create_model(self,name):
    """Create the model space with the file name 'name'"""
    do_real=SEP.par.logic(self.param("image_real"))
    return self._create_file(name,["mx","my","hx","hy","z_"],do_real)
   

#######################################################
#######################################################
#######################################################
class mig_off_par(off_par):
  """Class for building parameters for running migration"""
     
  def __init__(self,name="Offset migration"):
     """Initialize the migration parameters with a parameter database"""
     off_par.__init__(self,name)
                                                                                           
  def add_doc_params(self):
    """Add documentation specific to WEI  migration """
    off_par.add_doc_params(self)
    self.add_doc_param("az__n",doc="Number of depth steps")
    self.add_doc_param("az__o",doc="First depth")
    self.add_doc_param("az__d",doc="Sampling in depth")
    self.add_doc_param("jmx_image",'1',"Output every jth gather in midpoint x")
    self.add_doc_param("jmy_image",'1',"Output ecery jth gather in midpoint y")
    self.add_doc_param("image_real",'y',"whether or not the image is real")
    self.add_doc_param("wave_out",doc="File name for the wavefield at bottom of migration domain",required=None)
    self.add_doc_param("zero_image",doc="Zero-offset image",required=None)
                                                                                           
  def build_check_params(self):
     """Build the parameters for migration"""
     self.add_mid("D")
     self.add_off("D")
     self.set_rsep(self.create_model("junk"))
     off_par.build_check_params(self)


#######################################################
#######################################################
#######################################################
class ph_par(off_par):
   """Class for parameters when doing WEI with ph gathers"""
   def __init__(self,name="Angle WEI"):
     """Initialize WEI ph parameter database"""
     off_par.__init__(self,name)
   def build_check_params(self):
     """Check parameters when doing angle WEI"""
     off_par.build_check_params(self)
     R=self.rsep()
     self._axis_add(R,"px",3)
     self._axis_add(R,"py",4)

   def create_model(self,name):
     """Create the model space with the file name 'name'"""
     do_real=SEP.par.logic(self.param("image_real"))
     return self._create_file(name,["mx","my","px","py","z_"],do_real)


#######################################################
#######################################################
#######################################################
class mig_ph_par(ph_par):
  """Class for building parameters for running migration with ph gathers"""
                                                                                           
  def __init__(self,name="Angle migration"):
     """Initialize the migration parameters with a parameter database"""
     ph_par.__init__(self,name)
                                                                                           
  def add_doc_params(self):
    """Add documentation specific to WEI  migration """
    ph_par.add_doc_params(self)
    self.add_doc_param("apx_o",0.,"First ray parameter in",required=None)
    self.add_doc_param("apx_d",.0075,"Sampling in ray parameter in x")
    self.add_doc_param("apy_o",0.,"First ray parameter in",required=None)
    self.add_doc_param("apy_d",.0075,"Sampling in ray parameter in x")
    self.add_doc_param("R",doc="Image (mx,my,px,py,w)")
    self.add_doc_param("image_real",'y',"whether or not the image is real")
                                                                                           
  def build_check_params(self):
     """Build the parameters for migration"""
     self.add_mid("D")
     self.add_off("D")
     self.add_param("apy_n",self.param("ahy_n"))
     self.add_param("apx_n",self.param("ahx_n"))
     self.set_rsep(self.create_model('junk'))
     ph_par.build_check_params(self)


#######################################################
#######################################################
#######################################################
class mod_off_par(off_par):
  """Class for building parameters for running modelling"""
                                                                                           
  def __init__(self,name="Offset modeling"):
     """Initialize the modeling parameters with a parameter database"""
     off_par.__init__(self,name)
                                                                                           
  def add_doc_params(self):
    """Add documentation specific to WEI  migration """
    off_par.add_doc_params(self)
    self.add_doc_param("aw__n",doc="Number of frequency steps")
    self.add_doc_param("aw__o",doc="First frequency")
    self.add_doc_param("aw__d",doc="Sampling in frequency")
                                                                                           
  def build_check_params(self):
     """Build the parameters for modeling"""
     self.add_mid("R")
     self.add_off("R")
     self.set_dsep(self.create_data('junk'))
     off_par.build_check_params(self)

#######################################################
#######################################################
#######################################################
class mod_ph_par(ph_par):
  """Class for building parameters for running modelling with ph gathers"""
                                                                                           
  def __init__(self,name="Offset modeling"):
     """Initialize the modeling parameters with a parameter database"""
     ph_par.__init__(self,name)
                                                                                           
  def add_doc_params(self):
    """Add documentation specific to WEI  migration """
    par.add_doc_params(self)
    self.add_doc_param("aw__n",doc="Number of frequency steps")
    self.add_doc_param("aw__o",doc="First frequency")
    self.add_doc_param("aw__d",doc="Sampling in frequency")
    self.add_doc_param("ahx_o",doc="First sample in  half offset x")
    self.add_doc_param("ahx_d",doc="Sampling in half offset x")
    self.add_doc_param("ahy_o",doc="First sample in half offset y")
    self.add_doc_param("ahy_d",doc="Sampling in half offset y")
    self.add_doc_param("R",doc="Image (mx,my,px,py,w)")
                                                                                           
  def build_check_params(self):
     """Build the parameters for migration"""
     self.set_dsep(self.create_data('junkl'))
     self.add_mid("R")
     self.add_ph("R")
     ph_par.build_check_params(self)
     self.add_param("ray_par","y")




#######################################################
#######################################################
#######################################################
 


class wei_parfiles:
  """Class to build parallel files for WEI"""
  def __init__(self):
    """Initialize parallel file builders"""
    
  
  def make_parallel_files(self,domain_vector,range_vector,wei_pars,par_pars):
    """Make the parallel files for the WEI operation"""
    parfiles={}
    self.p=par_pars
    self.w=wei_pars
    nsect=int(self.param("nsect"))
    parfiles=self.make_par_schoice(parfiles,nsect)
    parfiles=self.make_par_slow(parfiles,nsect)
    parfiles=self.make_par_data(parfiles,nsect)
    if self.w.param("DOWN",error=None): 
      parfiles=self.make_par_datum(parfiles,nsect)
    if self.w.param("R",error=None): 
      parfiles=self.make_par_image(parfiles,nsect)
    if self.w.param("wave_out",error=None): 
      parfiles=self.make_par_wave(parfiles,nsect)
    if self.w.param("zero_image",error=None): 
      parfiles=self.make_par_zero(parfiles,nsect)
    self.parfiles=parfiles
    return parfiles
    
  def make_par_schoice(self,parfiles,nsect):
    """Make the parallel reference slowness file"""
    parfiles["schoice"]=SEP.parfile.copy(name=self.w.param("schoice"),
      tag="schoice=",isect_dict=SEP.parjob.simple_section_list(nsect),
        usage="INPUT",type="COPY",logdir=self.p.param("logdir"))
    return parfiles

  def make_par_slow(self,parfiles,nsect):
     """Make the parallel slowness file"""
     parfiles["slow"]=SEP.parfile.copy(name=self.w.param("S"),tag="S=",
        isect_dict=SEP.parjob.simple_section_list(nsect),
        usage="INPUT",type="COPY",logdir=self.p.param("logdir"))
     return parfiles

  def make_par_datum(self,parfiles,nsect):
     """Make the parallel data file"""
     parfiles["domain"]=SEP.parfile.distribute(name=self.w.param("DOWN"),
      tag="DOWN=", isect_dict=SEP.parjob.simple_section_list(nsect),
       space_only=1, dff_axis=5,logdir=self.p.param("logdir"), nblock=[nsect])
     return parfiles

  def make_par_data(self,parfiles,nsect):
     """Make the parallel data file"""
     parfiles["range"]=SEP.parfile.distribute(name=self.w.param("D"),tag="D=",
      isect_dict=SEP.parjob.simple_section_list(nsect),space_only=1,
       dff_axis=5,logdir=self.p.param("logdir"), nblock=[nsect])
     return parfiles

  def make_par_image(self,parfiles,nsect):
     """Make the parallel image file"""
     nfs=SEP.par.logic(self.p.param("nfs_image",error=None))
     if nfs:
       sz=int(self.w.param("az__n"))*4
       sz=sz*int(self.w.param("amx_n"))
       sz=sz*int(self.w.param("amy_n"))
       sz=sz*int(self.w.param("ahx_n"))
       sz=sz*int(self.w.param("ahy_n"))
       sz=sz/2048/1024/1024
       extra_args=""
       if sz > 1: extra_args=" nfiles=%d maxsize=2000 "%(sz+1)
       parfiles["domain"]=SEP.parfile.copy_share(name=self.w.param("R"),
        tag="R=", logdir=self.p.param("logdir"),
          isect_dict=SEP.parjob.simple_section_list(nsect)
         ,space_only=1,type="COPY",reuse_par="add=y",
         share_datapath=self.p.param("share_datapath"),
         nmax_procs=self.p.param("nmax_procs"),
         extra_args=extra_args)
     else:
       parfiles["domain"]=SEP.parfile.copy(name=self.w.param("R"),tag="R=",
         logdir=self.p.param("logdir"),
          isect_dict=SEP.parjob.simple_section_list(nsect)
         ,space_only=1,type="COPY",reuse_par="add=y")
     return parfiles


  def make_par_zero(self,parfiles,nsect):
     """Make the parallel wave out file"""
     parfiles["zero_image"]=SEP.parfile.copy(
       name=self.w.param("zero_image"),tag="zero_image=",
      isect_dict=SEP.parjob.simple_section_list(nsect),dff_axis=5,
      logdir=self.p.param("logdir"), usage="OUTPUT", nblock=nsect)
     return parfiles

  def make_par_wave(self,parfiles,nsect):
     """Make the parallel wave out file"""
     parfiles["wave_out"]=SEP.parfile.distribute(
       name=self.w.param("wave_out"),tag="wave_out=",
      isect_dict=SEP.parjob.simple_section_list(nsect),dff_axis=5,
      logdir=self.p.param("logdir"), usage="OUTPUT", nblock=nsect)
     return parfiles

  def sect_pars(self,data_vec,nsect):
    """Return the section paramters  for the jobs"""
    nw,ow,dw,label,unit=data_vec.axis(5)
    par_dict={}
    order=[]
    nw=int(nw)
    nsect=int(nsect)
    nb=int(int(nw)/int(nsect))
    nextra=nw-nsect*nb
    itot=0
    for i in range(nsect):
       key=str(i)
       order.append(key)
       par_dict[key]=SEP.par.sep_pars(name=str(i))
       o=ow+dw*itot
       n=nb
       if i < nextra: n=n+1
       par_dict[key].add_param("aw__n",n)
       par_dict[key].add_param("aw__o",o)
       par_dict[key].add_param("aw__d",dw)
       itot=itot+n
    return par_dict,order



#######################################################
#######################################################
#######################################################
 
class wpar_job_nfs(SEP.parjob.par_job_nfs,wei_parfiles):
  """Parallel object for a WEI job"""

  def __init__(self,name="PAR"):
    """Initialze a WEI para object"""
    SEP.parjob.par_job.__init__(self,name)
    wei_parfiles.__init__(self)

  def add_doc_params(self):
    """Add the parameters needed for doing a WEI parallel job"""
    SEP.parjob.par_job.add_doc_params(self)
    self.del_doc_param("global_pars")
    self.del_doc_param("sect_pars")
    self.del_doc_param("program")
    self.del_doc_param("files")
    self.del_doc_param("order")
    self.add_doc_param("nsect",doc="The number of jobs to run (defaults to nfreq)",
     required=None)
    self.add_doc_param("nmax_procs",8,"The number of processes that share an image file (only when nfs_image=y)")
    self.add_doc_param("share_datapath",SEP.datapath.datapath(),"Datapath for output shared files (only when nfs_image=y")

  def runtime_check_build(self,domain_vector,range_vector,wei_pars):
    """Build and check the parameters to the program

       domain_vector - The domain vector (image)
       range_vector  - The range vector (data)
       wei_pars      - The parameters for the WEI job

    """
    nsect=self.param("nsect")
    if not nsect: nsect=int(range_vector.history.param("n5"))
    self.add_param("nsect",nsect)
    par_dict,order=self.sect_pars(range_vector,nsect)

    self.add_param_obj("sect_pars",par_dict)
    self.add_param("nfs_image","y")
    self.add_param_obj("files",self.make_parallel_files(domain_vector,
     range_vector,wei_pars,self))
    self.add_param("program",wei_pars.param("program"))
    self.add_param_obj("order",order)
    SEP.parjob.par_job_nfs.runtime_check_build(self)



#######################################################
#######################################################
#######################################################
 
class wpar_job(SEP.parjob.par_job,wei_parfiles):
  """Parallel object for a WEI job"""

  def __init__(self,name="PAR"):
    """Initialze a WEI para object"""
    SEP.parjob.par_job.__init__(self,name)
    wei_parfiles.__init__(self)

  def add_doc_params(self):
    """Add the parameters needed for doing a WEI parallel job"""
    SEP.parjob.par_job.add_doc_params(self)
    self.del_doc_param("global_pars")
    self.del_doc_param("sect_pars")
    self.del_doc_param("program")
    self.del_doc_param("files")
    self.del_doc_param("order")
    self.add_doc_param("nsect",doc="The number of jobs to run (nfreq) ",
     required=None)

  def runtime_check_build(self,domain_vector,range_vector,wei_pars):
    """Build and check the parameters to the program

       domain_vector - The domain vector (image)
       range_vector  - The range vector (data)
       wei_pars      - The parameters for the WEI job

    """
    nsect=self.param("nsect")
    if not nsect: nsect=int(range_vector.history.param("n5"))
    self.add_param("nsect",nsect)
    par_dict,order=self.sect_pars(range_vector,nsect)
    self.add_param_obj("sect_pars",par_dict)
    self.add_param_obj("order",order)
    par_dict=SEP.parjob.simple_parameter_list(nsect)
    self.add_param_obj("files",self.make_parallel_files(domain_vector,
      range_vector,wei_pars,self))
    self.add_param("program",wei_pars.param("program"))
    SEP.parjob.par_job.runtime_check_build(self)




  
#######################################################
#######################################################
#######################################################
class wei_par(SEP.operator.oc_par,SEP.options.par_group):
  """A WEI parallel operator to migration"""
  def __init__(self,wei_par,par_par):
      """Initialize a parallel offset migration operator"""
      SEP.options.par_group.__init__(self,"WEI.PAR")
      self.par_par=par_par
      self.wei_par=wei_par
      self.add_doc_params()
      self.verb=None

  def set_verb(self,verb):
    self.verb=verb 
    
  def add_doc_params(self):
    """Add the required files to WEI operations"""
    self.add_doc_param("restart",doc="Whether or not we are restarting",required=None)
    self.add_doc_param("verb",doc="Whether or not to verbose when running operator",required=None)
  def build_check_params(self):
    """Check to see if the parameters make sense and build domain and range vectors"""
    val=self.param("restart")
    self.wei_par.add_param("restart",val)
    self.par_par.add_param("restart_com","restart=y")
    self.par_par.add_param("restart",val)
    self.wei_par.build_check_params()
    self.par_par.build_check_params()
  def runtime_check_build(self):
    """Check to see if the parameters make sense before a run"""
    dsep=self.wei_par.dsep()
    rsep=self.wei_par.rsep()
    self.par_par.runtime_check_build(rsep,dsep,self.wei_par)
    par=self.return_args().return_basic_pars()
    par.add_pars(self.wei_par.return_args().return_basic_pars())
    SEP.operator.oc_par.__init__(self,"WEI",rsep,dsep,self.par_par,msg="WEI",
     verb=self.verb,pars=par)
    self.wei_par.runtime_check_build()
    self.par_par.set_global_par_dict(self.wei_par.return_args())

  def migrate(self,restart):
    """Perform migration"""
    self.par_par.set_global_par("adj","y")
    self.par_par.update_parfile_dict("domain",
      self.par_par.parfiles["domain"].clone(name=self.wei_par.param("R"),
        usage="OUTPUT",restart=restart))
                                                                                
    self.par_par.update_parfile_dict("range",
      self.par_par.parfiles["range"].clone(name=self.wei_par.param("D"),
       usage="INPUT",restart=restart))
    self.par_par.run_job(restart)

  def model(self,restart):
    """Perform modelling"""
    self.par_par.set_global_par("adj","n")
    self.par_par.update_parfile_dict("domain",
      self.par_par.parfiles["domain"].clone(name=self.wei_par.param("R"),
       usage="INPUT",restart=restart))
                                                                                
    self.par_par.update_parfile_dict("range",
      self.par_par.parfiles["range"].clone(name=self.wei_par.param("D"),
       usage="OUTPUT",restart=restart))
    self.par_par.run_job(restart)

  def read_params(self,args,prefix="",delete=None):
    a,b=self.wei_par.read_params(args,prefix="%s%s"%(prefix,"wei_"),delete=delete)
    
    a2,b2=self.par_par.read_params(args,prefix="%s%s"%(prefix,"par_"),delete=delete)
    a3,b3=SEP.options.par_group.read_params(self,args,prefix)
    b.add_pars(b2)
    b.add_pars(b3)
    if a or a2 or a3: return 1,b
    return None,b
  def set_prog_options(self,opts):
    SEP.options.par_group.set_prog_options(self,opts)
    self.wei_par.set_prog_options(opts)
    self.par_par.set_prog_options(opts)
  def doc(self,prefix=""):
    d=self.wei_par.doc("%swei_"%prefix)
    d.extend(self.par_par.doc("%spar_"%prefix))
    return d
    
#######################################################
#######################################################
#######################################################
class wei_par_datum(wei_par):
  """A WEI parallel operator to do datuming"""
  def __init__(self,wei_obj,par_obj):
      """Initialize a parallel offset datuming operator"""
      wei_par.__init__(self,wei_obj,par_obj)
  def datum(self,restart):
    """Perform datuming"""
    self.par_par.set_global_par("adj","y")
    self.par_par.update_parfile_dict("domain",
      self.par_par.parfiles["domain"].clone(name=self.wei_par.param("DOWN"),
       usage="OUTPUT",restart=restart))
                                                                                
    self.par_par.update_parfile_dict("range",
      self.par_par.parfiles["range"].clone(name=self.wei_par.param("D"),
       usage="INPUT",restart=restart))
    self.par_par.run_job(restart)


  def runtime_check_build(self):
    """Check to see if the parameters make sense before a run"""
    dsep=self.wei_par.dsep()
    datumsep=self.wei_par.datumsep()
    val=str(self.param("verb"))
    self.par_par.runtime_check_build(datumsep,dsep,self.wei_par)
    par=self.return_args().return_basic_pars()
    par.add_pars(self.wei_par.return_args().return_basic_pars())
    SEP.operator.oc_par("WEI",datumsep,dsep,self.par_par,msg="WEI",
      pars=par,verb=val)
    self.wei_par.runtime_check_build()
    self.par_par.set_global_par_dict(self.wei_par.return_args())
